#
# Copyright 2020 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Components are a way to register objects using
    contract ids. A contract id is attached to an interface or a set 
    of interface.

    It is designed to have a general way about passing objects or behaviors to plugins
    or extension or other modules. It then enables for these modules or extensions to rely on the calling
    module behaviors without the need for these to do explicit imports
"""

import logging

from collections import namedtuple
from typing import Any, Callable

class ComponentManagerError(Exception):
    pass

class FactoryNotFoundError(ComponentManagerError):
    pass

class NoRegisteredFactoryError(ComponentManagerError):
    pass


LOGGER = logging.getLogger('SRVLOG')

FactoryEntry = namedtuple('FactoryEntry',('create_instance', 'service'))


class ComponentManager:

    def __init__(self) -> None:
        """ Component Manager
        """
        self._contractIDs = {}

    def register_entrypoints( self, category, *args, **kwargs ) -> None:
        """ Load extension modules 

            Loaded modules will do self-registration
        """
        from pkg_resources import iter_entry_points
        for ep in iter_entry_points(category):
            LOGGER.info("Loading module: %s:%s", category, ep.name)
            ep.load()(*args, **kwargs)

    def register_factory( self, contractID: str, factory: Callable[[],None] ) -> None:
        """ Register a factory for the given contract ID
        """
        if not callable(factory):
            raise ValueError('factory must be a callable object')

        LOGGER.debug("Registering factory: %s", contractID)
        self._contractIDs[contractID] = FactoryEntry(factory, None)

    def register_service( self, contractID: str, service: Any ) -> None:
        """ Register an instance object as singleton service
        """
        def nullFactory():
            raise NoRegisteredFactoryError(contractID)

        LOGGER.debug("Registering service: %s", contractID)
        self._contractIDs[contractID] = FactoryEntry( nullFactory , service )

    def create_instance( self, contractID: str ) -> Any:
        """ Create an instance of the object referenced by its
            contract id.
        """
        fe = self._contractIDs.get( contractID )
        if fe:
            return fe.create_instance()
        else:
            raise FactoryNotFoundError(contractID)

    def get_service( self, contractID: str ) -> Any:
        """ Return instance object as singleton
        """
        fe = self._contractIDs.get( contractID )
        if fe is None:
            raise FactoryNotFoundError(contractID)
        if fe.service is None:
            fe = fe._replace(service=fe.create_instance())
            self._contractIDs[contractID] = fe
        return fe.service


gComponentManager = ComponentManager()


def get_service( contractID: str ) -> Any:
    """ Alias to component_manager.get_service
    """
    return gComponentManager.get_service( contractID )


def create_instance( contractID: str ) -> Any:
    """ Alias to component_manager.create_instance
    """
    return gComponentManager.create_instance( contractID )


def register_entrypoints( category: str, *args, **kwargs ) -> None:
    """ Alias to component_manager.register_components
    """
    gComponentManager.register_entrypoints( category, *args, **kwargs )

#
# Declare factories or services with decorators
#

def register_service( contractID: str ) -> Any:
    def wrapper(obj: Any):
        gComponentManager.register_service(contractID, obj)
        return obj
    return wrapper


def register_factory( contractID: str ) -> Any:
    def wrapper(obj: Any):
        gComponentManager.register_factory(contractID, obj)
        return obj
    return wrapper





#
# Copyright 2021 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
    Wrapper around more tornado-like request handler for
    implementing Qgis server API
"""
import logging

from tornado.web import HTTPError # noqa F401
from typing import Optional 

from pyqgisserver.plugins import plugin_list, plugin_metadata, failed_plugins
from .handler import RequestHandler, register_handlers

from qgis.core import Qgis

LOGGER = logging.getLogger('SRVLOG')


class PluginCollection(RequestHandler):

    def get(self, name: Optional[str]=None) -> None: 
        """ Return plugin info
        """
        if name:
            if name in failed_plugins:
                self.write({ 'name': name, 'error_log' : failed_plugins[name], 'status': 'failed' })
                return
            # Return plugin informations
            metadata = plugin_metadata(name)
            if not metadata:
                raise HTTPError(404)
            self.write({'name': name, 'status': 'loaded', 'metadata': metadata })
        else:
            def link(name):
                return self.public_url(f"/{name}")
            # List all loaded plugins
            plugins = [{ 'name': n, 'status': 'loaded', 'link': link(n) } for n in plugin_list()]
            plugins.extend([{ 'name' : n, 'status': 'failed', 'link': link(n) } for n in failed_plugins])
            self.write({'plugins': plugins })

        
def register( serverIface ):
    """ Register plugins api handlers
    """
    # XXX See https://github.com/qgis/QGIS/issues/45439
    prefix = '/plugins' if Qgis.QGIS_VERSION_INT < 32200 else ''
    register_handlers(serverIface, "/plugins","PluginsManagment",
                      [
                          (rf'{prefix}/(?P<name>[^\/]+)/?$', PluginCollection),
                          (r'/', PluginCollection),
                      ])





#
# Copyright 2018 3liz
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
""" Logging handlers
"""
import logging

REQ_LOG_TEMPLATE = "{ip}\t{code}\t{method}\t{url}\t{time}\t{length}\t"
REQ_FORMAT = REQ_LOG_TEMPLATE+'{agent}\t{referer}'
RREQ_FORMAT = REQ_LOG_TEMPLATE

# Lies between warning and error
REQ = 21
RREQ = 22

LOGGER=logging.getLogger('SRVLOG')

def setup_log_handler(log_level, formatstr='%(asctime)s\t%(levelname)s\t[%(process)d]\t%(message)s',
                      stream = None ):
    """ Initialize log handler with the given log level
    """
    logging.addLevelName(REQ, "REQ")
    logging.addLevelName(RREQ, "RREQ")

    logger = LOGGER
    logger.setLevel(getattr(logging, log_level.upper()))
    # Init the root logger
    if not logger.handlers:
        channel = logging.StreamHandler(stream=stream)
        formatter = logging.Formatter(formatstr)
        channel.setFormatter(formatter)
        logger.addHandler(channel)
        return True
    return False

def format_log_request(handler):
    """ Format current request from the given tornado request handler

        :return a tuple (fmt,code,reqtime,length) where:
            fmt: the log string
            code: the http return code
            reqtime: the request time
            length: the size of the payload
    """
    request = handler.request
    code    = handler.get_status()
    reqtime = request.request_time()

    length  = handler._headers.get('Content-Length') or -1
    agent   = request.headers.get('User-Agent') or ""
    referer = request.headers.get('Referer')  or ""

    fmt = REQ_FORMAT.format(
        ip=request.remote_ip,
        method=request.method,
        url=request.uri,
        code=code,
        time=int(1000.0 * reqtime),
        length=length,
        referer=referer,
        agent=agent)

    return fmt, code, reqtime, length


def log_request(handler):
    """ Log the current request from the given tornado request handler

        :param handler: The request handler
        :param logger: an optional logger

        :return A tuple (code,reqtime,length) where:
            code: the http retudn code
            reqtime: the request time
            length: the size of the payload
    """
    fmt, code, reqtime, length = format_log_request(handler)
    LOGGER.log(REQ, fmt)
    return code, reqtime, length


def format_log_rrequest(path, code, method, query, reqtime, headers, addr=''):
    """ Format current r-request from the given response

        :param response: The response returned from the request
        :return A tuple (fmt,code,reqtime,length) where:
            fmt: the log string
            code: the http retudn code
            reqtime: the request time
            length: the size of the payload
    """
    length = -1
    try:
        length = headers['Content-Length']
    except KeyError:
        pass

    fmt = RREQ_FORMAT.format(
        ip=addr,
        method=method,
        url=f"{path.rstrip('/')}/{query}",
        code=code,
        time=int(1000.0 * reqtime),
        length=length)

    return fmt


def log_rrequest(*args, **kwargs):
    """ Log the current response request from the given response

        :return A tuple (code,reqtime,length) where:
            code: the http retudn code
            reqtime: the request time
            length: the size of the payload
    """
    fmt = format_log_rrequest(*args, **kwargs)
    LOGGER.log(RREQ, fmt)




#
# Copyright 2018 3liz
# Author: David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import tornado
import logging

from .basehandler import BaseHandler
from ..version import __version__
from ..config import config_to_dict

LOGGER=logging.getLogger('SRVLOG')


class PingHandler(BaseHandler):
    def get(self):
        self.write_json({ 'status': 'ok' })


class StatusHandler(BaseHandler):
    def get(self):

        stats = self.application.stats

        try:
            from qgis.core import Qgis
            QGIS_VERSION="{} ({})".format(Qgis.QGIS_VERSION_INT,Qgis.QGIS_RELEASE_NAME)
        except Exception as e:
            LOGGER.error("Cannot get Qgis version %s", e)
            QGIS_VERSION="n/a"

        response = dict(tornado_ver=tornado.version,
                        version = __version__,
                        author="3Liz",
                        author_url="http://3liz.com",
                        qgis_version=QGIS_VERSION,
                        documentation="http://{}/doc/".format(self.request.host),
                        configuration=config_to_dict(),
                        env=dict(os.environ))
                        
        response.update(stats.json())

        self.write_json(response)



#
# Copyright 2018 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Configuration management

Configuration can be done either by using aconfiguration file or with environnement variable.

Except stated otherwise, the rule for environnement variable names is ``QGSRV_<SECTION>_<KEY>`` all in uppercase.

Options 


"""

import os
import configparser
import logging
import functools

from typing import Any, Iterable, Tuple

from pyqgisservercontrib.core import componentmanager

getenv = os.getenv

LOGGER = logging.getLogger('SRVLOG')

CONFIG = configparser.ConfigParser(interpolation=configparser.ExtendedInterpolation())    
# Preserve case
CONFIG.optionxform = lambda opt: opt


def getenv2( env1, env2, default):
    """ Get value from alternate env variable
    """
    return getenv(env1,getenv(env2,default))


def print_config( fp ):
    """ print configuration to file
    """
    CONFIG.write(fp)


def load_configuration():
    """ Read configuration file

        Load server configuration from configuration file.
    """
    CONFIG.clear()

    LOGGER.info('loading configuration')

    CONFIG.add_section('server')
    CONFIG.set('server', 'port'          , getenv('QGSRV_SERVER_HTTP_PORT', '8080'))
    CONFIG.set('server', 'interfaces'    , getenv('QGSRV_SERVER_INTERFACES', '0.0.0.0'))
    CONFIG.set('server', 'workers'       , getenv('QGSRV_SERVER_WORKERS', '2'))
    CONFIG.set('server', 'timeout'       , getenv('QGSRV_SERVER_TIMEOUT', '20'))
    CONFIG.set('server', 'enable_filters', getenv('QGSRV_SERVER_ENABLE_FILTERS', 'yes'))
    CONFIG.set('server', 'http_proxy'    , getenv('QGSRV_SERVER_HTTP_PROXY', 'no'))
    CONFIG.set('server', 'proxy_url'     , getenv('QGSRV_SERVER_PROXY_URL' , ''))
    CONFIG.set('server', 'restartmon'    , getenv('QGSRV_SERVER_RESTARTMON' , ''))
    CONFIG.set('server', 'pluginpath'    , getenv('QGSRV_SERVER_PLUGINPATH' , ''))
    CONFIG.set('server', 'ssl'           , getenv('QGSRV_SERVER_SSL' , 'no'))
    CONFIG.set('server', 'ssl_cert'      , getenv('QGSRV_SERVER_SSL_CERT', ''))
    CONFIG.set('server', 'ssl_key'       , getenv('QGSRV_SERVER_SSL_KEY' , ''))
    CONFIG.set('server', 'cross_origin'  , getenv('QGSRV_SERVER_CROSS_ORIGIN' , 'yes'))
    CONFIG.set('server', 'status_page'   , getenv('QGSRV_SERVER_STATUS_PAGE'  , 'no'))
    CONFIG.set('server', 'allow_headers' , getenv('QGSRV_SERVER_ALLOW_HEADERS' , 'X-Qgis-,X-Lizmap-'))

    #
    # Logging
    #

    CONFIG.add_section('logging')
    CONFIG.set('logging', 'level', getenv('QGSRV_LOGGING_LEVEL', 'DEBUG'))
    CONFIG.set('logging', 'qgis_info', getenv('QGSRV_LOGGING_QGIS_INFO', 'no'))

    #
    # Api configuration
    #

    # Qgis api endpoints
    CONFIG.add_section('api.endpoints')
    CONFIG.set('api.endpoints', 'landing_page', getenv('QGSRV_API_ENDPOINTS_LANDING_PAGE','/ows/catalog'))

    # Services enabled
    CONFIG.add_section('api.enabled')
    CONFIG.set('api.enabled', 'landing_page', getenv('QGSRV_API_ENABLED_LANDING_PAGE','no'))

    # Landing page config mapping
    # see: https://github.com/qgis/QGIS/pull/38189#issuecomment-875682735
    CONFIG.add_section('api:landing_page')
    CONFIG.set('api:landing_page', 'QGIS_SERVER_LANDING_PAGE_PREFIX', 
               '${api.endpoints:landing_page}')
    CONFIG.set('api:landing_page', 'QGIS_SERVER_LANDING_PAGE_PROJECTS_DIRECTORIES', 
               '${projects.cache:rootdir}')

    #
    # Projects cache
    #
    CONFIG.add_section('projects.cache')
    CONFIG.set('projects.cache', 'size'    , getenv('QGSRV_CACHE_SIZE','10' ))
    CONFIG.set('projects.cache', 'rootdir' , getenv('QGSRV_CACHE_ROOTDIR',''))
    CONFIG.set('projects.cache', 'strict_check' , getenv('QGSRV_CACHE_STRICT_CHECK','yes'))
    CONFIG.set('projects.cache', 'insecure'     , getenv('QGSRV_CACHE_INSECURE','no'))
    CONFIG.set('projects.cache', 'preload_config'      , getenv('QGSRV_CACHE_PRELOAD_CONFIG',''))
    # Use same variable name as Qgis server options
    CONFIG.set('projects.cache', 'trust_layer_metadata', 
               getenv2('QGSRV_TRUST_LAYER_METADATA','QGIS_SERVER_TRUST_LAYER_METADATA','no'))
    CONFIG.set('projects.cache', 'disable_getprint'    , 
               getenv2('QGSRV_DISABLE_GETPRINT','QGIS_SERVER_DISABLE_GETPRINT','no'))
    CONFIG.set('projects.cache', 'disable_owsurls'    , getenv('QGSRV_CACHE_DISABLE_OWSURLS','no'))
    
    # 
    CONFIG.add_section('projects.schemes')

    #
    CONFIG.add_section('zmq')
    # Identity prefix used in 0MQ worker socket 
    CONFIG.set('zmq', 'identity'     , getenv('QGSRV_ZMQ_IDENTITY' ,'OWS-SERVER'))
    # Address to bind 0MQ socket - used only with proxy/worker configuration
    CONFIG.set('zmq', 'bindaddr'     , getenv('QGSRV_ZMQ_INADDR'   ,'tcp://*:18080'))
    # Control the maximum lenghth of the waiting queue
    CONFIG.set('zmq', 'maxqueue'     , getenv('QGSRV_ZMQ_MAXQUEUE' ,'1000'))
    # Control the lifetime of requests on the waiting queue 
    CONFIG.set('zmq', 'timeout'      , getenv('QGSRV_ZMQ_TIMEOUT'  ,'15000'))
    # Address to bind broadcast address to - used only with proxy/worker configuration
    CONFIG.set('zmq', 'broadcastaddr', getenv('QGSRV_ZMQ_BROADCASTADDR','tcp://*:18090'))

    #
    # Monitoring (AMQP)
    #
    CONFIG.add_section('monitor:amqp')
    CONFIG.set('monitor:amqp','routing_key', getenv('AMQP_ROUTING',''))
    CONFIG.set('monitor:amqp','host'       , getenv('AMQP_HOST','amqp'))
    CONFIG.set('monitor:amqp','user'       , getenv('AMQP_USER',''))
    CONFIG.set('monitor:amqp','vhost'      , getenv('AMQP_VHOST','/'))
    CONFIG.set('monitor:amqp','port'       , getenv('AMQP_PORT','5672'))
    CONFIG.set('monitor:amqp','exchange'   , getenv('AMQP_EXCHANGE','qgis_log'))
    # Beware that a too small reconnect delay 
    # may prevent other asynchronous tasks to run
    CONFIG.set('monitor:amqp','reconnect_delay', getenv('AMQP_RECONNECT_DELAY','5'))

    #
    # Management 
    #
    CONFIG.add_section('management')
    CONFIG.set('management','enabled'    , getenv('QGSRV_MANAGEMENT_ENABLED','no'))
    CONFIG.set('management','interfaces' , getenv('QGSRV_MANAGEMENT_INTERFACES', '127.0.0.1'))
    CONFIG.set('management','ssl'        , getenv('QGSRV_MANAGEMENT_SSL','no'))
    CONFIG.set('management','ss_key'     , getenv('QGSRV_MANAGEMENT_SSL_KEY' ,''))
    CONFIG.set('management','ssl_cert'   , getenv('QGSRV_MANAGEMENT_SSL_CERT',''))
    CONFIG.set('management','port'       , getenv('QGSRV_MANAGEMENT_PORT','19876'))
  

def read_config_file( cfgfile ):
    """ Read configuration from file
    """
    CONFIG.read_file(cfgfile)
    LOGGER.info('Configuration file <%s> loaded', cfgfile)


def config_to_dict():
    """ Convert actual configuration to dictionary
    """
    return { s: dict(p.items()) for s,p in CONFIG.items() }


def validate_config_path(confname, confid, optional=False):
    """ Validate directory path
    """
    confvalue = CONFIG[confname].get(confid,'')

    if not confvalue and optional:
        return

    confvalue = os.path.normpath(confvalue)
    if not os.path.isdir(confvalue):
        LOGGER.error('server->%s configuration value %s is not directory' % (confid, confvalue))
        raise ValueError(confvalue)

    if not os.path.isabs(confvalue):
        LOGGER.error('server->%s configuration value %s is not absolute path' % (confid, confvalue))
        raise ValueError(confvalue)

    CONFIG.set(confname, confid, confvalue)


def configure_qgis_api( name: str ) -> None:
    """ Configure qgis service environnement variables
    """
    config = CONFIG[f"api:{name}"]
    for k,v in config.items():
        LOGGER.debug("configuring qgis api '%s': %s = %s", name, k, v) 
        os.environ[k] = v


def qgis_api_endpoints(enabled_only: bool=True) -> Iterable[Tuple[str,str]]:
    """ Return the list of enabled services
    """
    endpoints = CONFIG["api.endpoints"]
    enabled   = CONFIG["api.enabled"] 
    items     = ((name,endpoint) for name,endpoint in endpoints.items())
    if enabled_only:
        items = filter(lambda item: enabled.getboolean(item[0]),items) 
    return items
        

#
# Published services
#

NO_DEFAULT=object()

# Chars that are to be replaced in env variable name
ENV_REPLACE_CHARS=':.-@#$%&*'

@componentmanager.register_factory('@3liz.org/config-service;1')
class ConfigService:
    """ Act as a proxy
    """ 

    def __init__(self):
        self.allow_env = True

    def __get_impl( self, _get_fun, section:str, option:str, fallback:Any = NO_DEFAULT ) -> Any:
        """
        """
        value = _get_fun(section, option, fallback=NO_DEFAULT)
        if value is NO_DEFAULT:
            # Look in environment
            # Note that the section must exists
            if self.allow_env:
                varname  = 'QGSRV_%s_%s' % (section.upper(), option.upper())
                varname  = functools.reduce( lambda s,c: s.replace(c,'_'), ENV_REPLACE_CHARS, varname)
                varvalue = os.getenv(varname)
                if varvalue is not None:
                    LOGGER.debug("Setting config value from %s", varname)
                    CONFIG.set(section, option, varvalue)
                # Let config parser translate the value for us
                value = _get_fun(section, option, fallback=fallback)
            else:
                value = fallback
        if value is NO_DEFAULT:
            raise KeyError('[%s] %s' % (section,option))
        return value

    get        = functools.partialmethod(__get_impl,CONFIG.get) 
    getint     = functools.partialmethod(__get_impl,CONFIG.getint) 
    getboolean = functools.partialmethod(__get_impl,CONFIG.getboolean) 
    getfloat   = functools.partialmethod(__get_impl,CONFIG.getfloat) 

    def __getitem__(self, section):
        return CONFIG[section]

    def __contains__(self, section):
        return section in CONFIG

    def set( self, section:str, option:str, value: Any ) -> None:
        CONFIG.set( section, option, value )

    def add_section( self, sectionname:str ) -> None:
        # We do not care if the section already exists
        try:
            CONFIG.add_section( sectionname )
        except configparser.DuplicateSectionError:
            pass



confservice = ConfigService()





""" AMQP monitor for qgis requests
"""
import asyncio
import logging
import traceback
import json
import os

from typing import Union, Dict, Optional

from .config  import confservice

LOGGER = logging.getLogger('SRVLOG')

def _decode( b: Union[str,bytes] ) -> str:
    if not isinstance(b,str):
        return b.decode('utf-8')
    return b

TAG_PREFIX = 'AMQP_GLOBAL_TAG_' 


def _read_credentials( vhost: str, user: str ) -> Optional:  # ['PlainCredentials']
    """ Read credentials from passfile
    """
    credential_file = os.getenv("AMQPPASSFILE")
    if not (credential_file and os.path.exists(credential_file)):
        return

    from pika import PlainCredentials

    LOGGER.debug("Using passfile %s", credential_file)
    with open(credential_file) as fp:
        for line in fp.readlines():
            credentials = line.strip()
            if credentials and not credentials.startswith('#'):
                cr_vhost, cr_user, passwd = credentials.split(':')
                if cr_vhost in ('*',vhost) and  cr_user in ('*',user):
                    LOGGER.info("Using credentials for user '%s' on vhost '%s'", user, vhost)
                    return PlainCredentials(user,passwd)


class Monitor:

    def __init__(self, amqp_client: 'AsyncPublisher', routing_key: str ) -> None: # noqa: F821
        """ Init AMQP monitor
        """
        self._client = amqp_client

        self._dynamic_routing = routing_key.startswith('@')
        if self._dynamic_routing:
            self._routing_key = routing_key[1:]
        else:
            self._routing_key = routing_key

        # Get global tags
        tags = ((e.partition(TAG_PREFIX)[2],os.environ[e]) for e in os.environ if e.startswith(TAG_PREFIX))
        self._global_tags = { t:v for (t,v) in tags if t }
    

    def emit( self, status:int, arguments: Dict, delta: float, meta: Dict ) -> None:
        """ Publish monitor data
        """
        if self._dynamic_routing:
            routing_key = self._routing_key.format(META=meta)
        else:
            routing_key = self._routing_key

        params = { k:_decode(v[0]) for k,v in arguments.items() }
        # Send all params to our logger
        ms = int(delta * 1000.0)
        params.update(self._global_tags,
                      RESPONSE_TIME=ms,
                      RESPONSE_STATUS=status,
                      ROUTING_KEY=routing_key)
        log_msg = json.dumps(params)
        self._client.publish( log_msg ,
                              routing_key  = routing_key,
                              expiration   = 3000,
                              content_type = 'application/json',
                              content_encoding ='utf-8')

    @classmethod
    def initialize(cls) -> 'Monitor':
        """ Register an instance of Monitor client
        """
        conf = confservice['monitor:amqp']
        routing_key = conf.get('routing_key')
        if not routing_key:
            return 

        try:
            from amqpclient.concurrent import AsyncPublisher
        except ImportError:
            LOGGER.warning("Cannot import 'amqpclient', AMQP logging will not be available")
            return None

        hosts = conf['host']
        user  = conf['user']
        vhost = conf['vhost']
        port  = conf['port']

        reconnect_delay = conf.getint('reconnect_delay')

        kwargs = {}

        if user:
            credentials = _read_credentials( vhost, user )
            if credentials:
                kwargs['credentials'] = credentials

        exchange = conf['exchange']

        client = AsyncPublisher(host=hosts,port=int(port),virtual_host=vhost,
                                reconnect_delay=reconnect_delay,
                                logger=LOGGER, **kwargs)

        # Catch exception in connection
        async def connect():
            try:
                await client.connect(exchange=exchange,exchange_type='topic')
                LOGGER.info("AMQP logger initialized.")
            except Exception:
                LOGGER.error("Failed to initialize AMQP logger: %s",traceback.format_exc())

        asyncio.ensure_future( connect() )

        inst = cls(client, routing_key)
        setattr(cls,'_instance', inst)
        return inst




#
# Copyright 2020 3liz
# Author: David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
"""
The fork serve will ensure that forking processes 
occurs from [almost] the same state.
"""
import os
import zmq
import logging
import signal
import time
import traceback
import asyncio

from glob import glob

from multiprocessing import Process
from multiprocessing.util import Finalize

from typing import Callable, Awaitable, Optional, Dict, List

from .zeromq.supervisor import Supervisor
from .zeromq.pool import Pool

from .config import confservice

from .qgsworker import QgsRequestHandler

from pyqgisservercontrib.core.watchfiles import watchfiles

LOGGER=logging.getLogger('SRVLOG')

class _RestartHandler:

    def __init__(self) -> None:
        self._restart = None
        self._watch_files = []

    def update_files(self) -> None:
        """ update files to watch
        """
        conf = confservice['server']

        self._watch_files.clear()
        restartmon = conf.get('restartmon')
        if restartmon:
            self._watch_files.append(restartmon)

        # Check for plugins
        pluginpath = conf.get('pluginpath')
        if pluginpath:
            plugins = glob(os.path.join(pluginpath,'*/.update-manifest'))
            self._watch_files.extend(plugins)

        LOGGER.debug("Updated watch files %s", self._watch_files)

    def close(self) -> None:
        if self._restart:
            self._restart.stop()

    def start(self, do_restart: Callable[[None],None]) -> None:
        """ Create a restart handler
        """
        self.update_files()
        
        def callback( *args ):
            do_restart()
            self.update_files()

        check_time = confservice.getint('server','restartmon_check_time', 3000)
        self._restart = watchfiles(self._watch_files, callback, check_time)
        self._restart.start()


class _Server:

    def __init__(self, broadcastaddr: str, pool: Process,  timeout: int,
                 num_workers: int) -> None:

        ctx = zmq.Context.instance()
        pub = ctx.socket(zmq.PUB)
        pub.setsockopt(zmq.LINGER, 500)    # Needed for socket no to wait on close
        pub.setsockopt(zmq.SNDHWM, 1)      # Max 1 item on send queue
        pub.bind(broadcastaddr)

        self._timeout = timeout 
        self._sock = pub
        self._num_workers = num_workers

        LOGGER.debug("Started pool server")
        self._pool = pool
        self._supervisor = None
        self._healthcheck = None

        self._restart_handler = _RestartHandler()
        self._restart_handler.start(self.restart)

        # Ensure that pool is terminated is called
        # at process exit
        self._terminate = Finalize(
            self, self._terminate_pool, 
            args=(self._pool,),
            exitpriority=16
        )

    async def healthcheck(self) -> Awaitable[None]:
        while True:
            if self._pool.exitcode is not None and self._pool.exitcode != 0:
                LOGGER.critical("Pool failure, exiting...")
                raise SystemExit("Exiting because of unrecoverable error")
            await asyncio.sleep(5)
    
    def start_supervisor(self):
        """ Start supervisor independently

            Note: It is no recommended to run supervisor before asyncio loop
            has been properly set - for exemple when using a custom loop.
        """
        if self._supervisor is None:
            LOGGER.info("Initializing supervisor")
            self._supervisor = Supervisor(self._timeout)
            self._supervisor.run()

        if self._healthcheck is None:
            LOGGER.info("Initializing pool healthcheck")
            self._healthcheck = asyncio.ensure_future(self.healthcheck())

    @classmethod
    def _terminate_pool(cls, p: Process) -> None:
        if p and hasattr(p, 'terminate'):
            if p.exitcode is None:
                p.terminate()
            if p.is_alive():
                p.join()

    def terminate(self):
        """ Terminate handler
        """
        LOGGER.info("Stopping pool server")
        self._restart_handler.close()
        if self._healthcheck:
            self._healthcheck.cancel()
        self._sock.close()
        if self._supervisor:
            self._supervisor.stop()
        LOGGER.info("Stopping worker pool")
        self._terminate()

    def broadcast(self, command: bytes ) -> None:
        """ Broadcast notification to workers 
        """
        try:
            self._sock.send(command, zmq.NOBLOCK)
        except zmq.ZMQError as err:
            if err.errno != zmq.EAGAIN:
                LOGGER.error("Broadcast Error %s\n%s", err, traceback.format_exc())

    def restart(self) -> None:
        """ Send restart command
        """
        self.broadcast(b'RESTART')

    @property
    def num_workers(self) -> int:
        return self._num_workers

    async def get_reports(self) -> Awaitable[List[Dict]]:
        """ Collect reports
        """
        maxwait=10
        so_far=0
        self._supervisor.clear_reports()
        self.broadcast(b'REPORT')
        while self._supervisor.num_reports() < self._num_workers:
            await asyncio.sleep(1)
            so_far += 1
            if so_far >= maxwait:
                break
        return self._supervisor.reports


def create_poolserver(numworkers: int) -> _Server:
    """ Run workers pool in its own process

        This ensure that sub-processes all always forked from
        the same parent context
    """
    router        = confservice['zmq']['bindaddr']
    broadcastaddr = confservice['zmq']['broadcastaddr']
    timeout       = confservice['server'].getint('timeout')

    maxcycles = confservice.getint('server', 'maxcycles', fallback=None)
    if maxcycles is not None:
        if maxcycles <= 0:
            maxcycles = None
        else:
            LOGGER.info(f"Max cycles limit set to {maxcycles}")

    p = Process(target=run_worker_pool, args=(numworkers, broadcastaddr, router, maxcycles))
    p.start()

    poolserver = _Server(broadcastaddr, p, timeout, numworkers)
    return poolserver


def run_worker_pool(numworkers: int, broadcastaddr: str, router: str, 
                    maxcycles: Optional[int]=None) -> None:
    """ Run a qgis worker pool

        Ensure that child processes run in the main thread
    """

    # Try to exit gracefully
    def term_signal(signum,frames):
        #print("Caught signal: %s" % signum, file=sys.stderr)
        raise SystemExit()

    LOGGER.info("Starting worker pool")
    pool = Pool( numworkers, target=QgsRequestHandler.run, args=(router,),
                 kwargs={ 'broadcastaddr': broadcastaddr, 'maxcycles': maxcycles } )

    signal.signal(signal.SIGTERM,term_signal)

    try:
        while True:
            if pool.critical_failure:
                raise RuntimeError("Server aborting prematurely !")
            pool.maintain_pool()
            time.sleep(0.1)
    except (KeyboardInterrupt,SystemExit):
        LOGGER.warning("Pool Interrupted")
    finally:
        LOGGER.info("Terminating worker pool")
        pool.terminate()





#
# Copyright 2018 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Load balancer/proxy to morker processes

    Implement ZMQ ROUTER/ROUTER broker for connecting client DEALER 
    to workers DEALER
"""

import sys
import zmq
import logging
import traceback
import signal

from time import time
from collections import deque

from .messages import WORKER_READY

from ..logger import setup_log_handler

LOGGER=logging.getLogger('SRVLOG')


def run_broker( inaddr: str, outaddr: str, maxqueue: int=100, timeout: int=3000) -> None:
    """ Create a ROUTER-ROUTER broker

        :param inaddr: frontend address to bind to
        :param outaddr: backend address to bind to

        If the max number of waiting request is reached: extra incoming requests will
    """
    # Convert timeout to seconds
    timeout = timeout/1000.0

    context = zmq.Context.instance()

    LOGGER.info("Binding frontend to %s", inaddr)
    frontend = context.socket(zmq.ROUTER)
    frontend.setsockopt(zmq.LINGER, 500)
    frontend.setsockopt(zmq.ROUTER_MANDATORY,1)
    frontend.bind(inaddr)

    LOGGER.info("Binding backend to %s", outaddr)
    backend = context.socket(zmq.ROUTER)
    backend.setsockopt(zmq.LINGER, 500)
    backend.setsockopt(zmq.ROUTER_MANDATORY,1)
    backend.bind(outaddr)

    poller = zmq.Poller()    

    # Only poll for requests from backend until workers are available
    poller.register(backend,  zmq.POLLIN)
    poller.register(frontend, zmq.POLLIN)
   
    workers = set() # Workers available
    waiting = deque() # Client waiting

    LOGGER.info("Starting ZMQ broker loop")

    # Try to exit gracefully
    def term_signal(signum,frames):
        LOGGER.debug("Broker process caught signal: %s" % signum)
        raise SystemExit()

    signal.signal(signal.SIGTERM,term_signal)

    try:
        while True:
            # Poll incoming requests
            sockets = dict(poller.poll())    

            # Handle worker activity on the backends
            if backend in sockets:
                try:
                    worker_id, client_id, *rest = backend.recv_multipart()

                    if client_id == WORKER_READY:
                        # Worker is available on new connection
                        # Mark worker as available
                        if worker_id not in workers:
                            LOGGER.debug("READY %s", worker_id)
                            workers.add(worker_id)
                    else:
                        msgid, data = rest
                        try:
                            frontend.send_multipart([client_id, msgid, data])
                            LOGGER.debug("SND worker: %s -> client: %s : %s", worker_id, client_id, msgid)
                        except zmq.ZMQError as err:
                            # ZMQ Will raise error if no client_id connected
                            LOGGER.error("SND worker: %s -> client: %s', %s ,errno %s", worker_id, client_id, err, err.errno)
                except Exception:
                    LOGGER.error("%s", traceback.format_exc())

            # Handle incoming client requests
            if frontend in sockets:
                try:
                    client_id, msgid, data = frontend.recv_multipart()
                    LOGGER.debug("REQUEST %s %s", client_id, msgid)
                    # Push on waiting queue
                    if len(waiting) >= maxqueue:
                        LOGGER.error("Max waiting requests reached (max %d)", maxqueue)
                        try:
                            frontend.send_multipart([client_id, msgid, b"ERR", b"509"])
                        except zmq.ZMQError as err:
                            LOGGER.error("SND ERR -> client: %s, %s, errno %s", client_id, err, err.errno)
                    else:
                        waiting.appendleft((time(), client_id, msgid, data))
                except Exception:
                    LOGGER.error("%s", traceback.format_exc())

            # Handle waiting requests
            # Unavailable workers will be automatically removed from the list
            if workers and waiting:
                now = time()
                while workers and waiting:
                    tm, client_id, msgid, data = waiting.pop()
                    # Test timeout
                    if now - tm > timeout:
                        LOGGER.debug("DROP %s: %s", client_id, msgid)
                        continue
                    while workers:
                        worker_id = workers.pop()
                        try:
                            backend.send_multipart([worker_id, client_id, msgid, data])
                            LOGGER.debug("SND client: %s -> worker: %s : %s", client_id, worker_id, msgid)
                            break # Handle next request
                        except zmq.ZMQError as err:
                            LOGGER.info("SND client: %s -> worker: %s, %s, errno %s", client_id, worker_id, err, err.errno)
                            if not workers: 
                                # No more workers available
                                # push back the request on the queue
                                waiting.append((tm, client_id, msgid, data))
                            
    except (KeyboardInterrupt, SystemExit):
        LOGGER.warning("Broker Terminated")
    except Exception:
        LOGGER.critical("Uncaught Exception:\n%s", traceback.format_exc())
    finally:
        backend.close()
        frontend.close()
        context.term()

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description='Test broker')
    parser.add_argument('--iface', metavar="host", default="tcp://127.0.0.1", help="Interface to bind to")
    parser.add_argument('--in'   , dest='inaddr' , metavar='address', default='{iface}:8880', help="frontend address")
    parser.add_argument('--out'  , dest='outaddr', metavar='address', default='{iface}:8881', help="backend address")
    parser.add_argument('--logging', choices=['debug', 'info', 'warning', 'error'], default='info', help="set log level")
    parser.add_argument('--maxqueue', metavar='NUM', type=int, default=100, help="Max waiting queue")
    parser.add_argument('--timeout' , metavar='NUM', type=int, default=3000, help="Set timeout in ms for waiting requests")

    args = parser.parse_args()

    setup_log_handler(args.logging)
    print("Log level set to {}\n".format(logging.getLevelName(LOGGER.level)), file=sys.stderr)

    LOGGER.setLevel(getattr(logging, args.logging.upper()))

    run_broker(args.inaddr.format(iface=args.iface), 
               args.outaddr.format(iface=args.iface),
               maxqueue=args.maxqueue,
               timeout=args.timeout)
    print("DONE", file=sys.stderr)



#
# Copyright 2020 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Pool server 
"""
import os
import logging
import time
import signal

from multiprocessing import Process
from multiprocessing.util import Finalize

from typing import Callable,List,Dict

# Early failure min delay
# If any process fail before that starting delay
# we abort the whole process
EARLY_FAILURE_DELAY = 5

LOGGER = logging.getLogger('SRVLOG')


class Pool:

    def __init__(self, num_workers: int, target: Callable, args: List=(), 
                 kwargs: Dict={}) -> None:

        self.critical_failure = False

        self._num_workers = num_workers
        self._pool = []
        self._args = args
        self._kwargs = kwargs
        self._target = target
        self._start_time = time.time()

        # Ensure that pool is terminated is called
        # at process exit
        self._terminate = Finalize(
            self, self._terminate_pool, 
            args=(self._pool,),
            exitpriority=15
        )

        self._repopulate_pool()

    def _join_exited_workers(self) -> bool:
        """Cleanup after any worker processes which have exited due to reaching
           their specified lifetime.  
           
           Returns True if any workers were cleaned up.
        """
        cleaned = False
        for i in reversed(range(len(self._pool))):
            worker = self._pool[i]
            if worker.exitcode is not None:
                # Handle special case when 
                # Return is not zero
                if worker.exitcode != 0:
                    # Handle early failure by killing current process
                    LOGGER.warning("Worker %s exited with code %s", worker.pid, worker.exitcode) 
                    if time.time() - self._start_time < EARLY_FAILURE_DELAY:
                        # Critical exit
                        LOGGER.critical("Critical worker failure. Aborting...")
                        self.critical_failure = True
                        os.kill(os.getpid(), signal.SIGABRT)
                # worker exited
                worker.join()
                cleaned = True
                del self._pool[i]
        return cleaned

    def _repopulate_pool(self) -> None:
        """Bring the number of pool processes up to the specified number,
        for use after reaping workers which have exited.
        """
        for _ in range(self._num_workers - len(self._pool)):
            w = Process(target=self._target, args=self._args, kwargs=self._kwargs)
            self._pool.append(w)
            w.name = w.name.replace('Process', 'PoolWorker')
            w.start()

    def maintain_pool(self) -> None:
        """Clean up any exited workers and start replacements for them.
        """
        if self._join_exited_workers():
            self._repopulate_pool()

    @classmethod
    def _terminate_pool(cls, pool: 'Pool') -> None: 

        # Send terminate to workers
        if pool and hasattr(pool[0], 'terminate'):
            for p in pool:
                if p.exitcode is None:
                    p.terminate()

        # Join pool workers
        if pool and hasattr(pool[0], 'terminate'):
            for p in pool:
                if p.is_alive():
                    # worker has not yet exited
                    p.join()

    def __reduce__(self) -> None:
        raise NotImplementedError(
            'Pool objects cannot be passed between processes or pickled'
        )

    def terminate(self) -> None:
        self._terminate()

    def kill(self, pid: int) -> bool:
        """ Kill a running process
        """
        try:
            p = next(filter(lambda w: w.pid == pid, self._pool))
            p.kill()
            return True
        except StopIteration:
            pass
        return False




#
# Copyright 2021 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from typing import Dict, Optional

try:
    import psutil
    HAVE_PSUTIL = True
except Exception:
    print("WARNING: Failed to load 'PSutil', system metrics will not be collected")
    HAVE_PSUTIL = False


def stats(pid: Optional[int]=None) -> Dict:
    """ Collect stats about process
        see https://psutil.readthedocs.io/en/latest/#processes
    """
    if not HAVE_PSUTIL:
        return {}

    proc = psutil.Process(pid)
    return dict(
        mem_usage=proc.memory_info().rss,
        mem_percent=proc.memory_percent(),
    )    



#
# Copyright 2018 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Define various decarators for:

    singleton,
    lazy properties,
    cached function evaluation
"""

_missing = object()

__all__ = ['singleton', 'once', 'lazy_property']


class singleton:
    """ Decorator for defining a singleton-like object instanciated at the first call """
    def __init__(self, decorated):
        self._decorated = decorated

    def __call__(self, *args, **kwargs):
        try:
            return self._instance
        except AttributeError:
            self._instance = self._decorated(*args, **kwargs)
            return self._instance


class once:
    """ Evaluate the function only once """
    def __init__(self, f):
        self.f = f

    def __call__(self, *args, **kwargs):
        if not hasattr(self, "value"):
            self.value = self.f(*args, **kwargs)
        return self.value


class lazy_property:
    """ A decorator that converts a function into a lazy property. The
        function wrapped is called the first time to retrieve the result
        and then that calculated result is used the next time you access
        the value::

            class Foo(object):

            @lazy_property
            def foo(self):
                # calculate something important here
                return 42

        The class has to have a `__dict__` in order for this property to
        work.
    """

    # implementation detail: this property is implemented as non-data
    # descriptor. non-data descriptors are only invoked if there is
    # no entry with the same name in the instance's __dict__.
    # this allows us to completely get rid of the access function call
    # overhead. If one choses to invoke __get__ by hand the property
    # will still work as expected because the lookup logic is replicated
    # in __get__ for manual invocation.

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __get__(self, obj, klass=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, _missing)
        if value is _missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value



#
# Copyright 2021 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
    Collect global stats
"""
from time import time
from datetime import datetime, timedelta

class Stats:

    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        self.num_requests = 0
        self.num_errors   = 0
        self.start_time   = time()
       
    def json(self):
        """ Return a json payload
        """
        return dict(
            num_request = self.num_requests,
            num_errors  = self.num_errors,
            start_date  = datetime.fromtimestamp(self.start_time).isoformat(),
            uptime      = timedelta(seconds=time() - self.start_time).total_seconds(),
        )




#
# Copyright 2018 3liz
# Author: David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Define Qgis Worker

    We define QGis workers as DEALER configuration: it will enables
    us to forward streamed response as chunked transfer.  

    1. Start qgis application
    2. Connect a REQ zmq socket to endpoint
    3. Return qgis server responses
"""
import os
import sys
import logging
import traceback
import zmq
import pickle
import uuid

from typing import TypeVar, Optional, Type

from ..logger import setup_log_handler
from ..utils import stats

from .messages import (WORKER_READY, ReplyMessage)
from .supervisor import Client as SupervisorClient



LOGGER=logging.getLogger('SRVLOG')

# Define an abstract type for HTTPRequest
HTTPRequest = TypeVar('HTTPRequest')


class RequestHandler:

    def __init__(self, socket: zmq.Socket, client_id: bytes, correlation_id: bytes, request: HTTPRequest) -> None:
        """ Handle requests and 
        
            Handle reply message contruction and pass message correlation_id to 
            reply.

            :param request: An HTTP request handler
            :param socket: the zmq socket
        """
        self.headers = {}
        self.status_code = 200
        self.request = request
        self.header_written = False

        self._correlation_id = correlation_id
        self._socket    = socket
        self._client_id = client_id

    def _write( self, data: bytes ) -> None:
        """ Send data back to client
        """
        self._socket.send_multipart([
            self._client_id,
            self._correlation_id,
            data])

    def send( self, data: bytes, send_more: bool=False ) -> None:
        """ Send data
        """
        if not self.header_written:
            # We let the client know that there is more 
            # data by setting the 206 code (partial response)
            if send_more and self.status_code==200:
                self.status_code = 206
            # Create a Header Message
            message = pickle.dumps(ReplyMessage(self.status_code, headers=self.headers, data=data),-1)
            self._write( message ) 
            self.header_written = True
        elif self.status_code == 206:
            self._write(pickle.dumps((data, send_more),-1))
            if not send_more:
                self.status_code = 200

    @property
    def msgid(self):
        """ Return the message correlation id
        """
        return self._correlation_id

    @property
    def identity(self) -> bytes:
        return self._socket.identity

    def handle_message(self) -> None:
        """ Override this method to handle_messages
        """
        print("Received", self.request.data, "from", self._client_id)
        self.send(b"Hello %s" % self._client_id, True)
        self.send(b"Chunk 1", True)
        self.send(b"Chunk 2", True)
        self.send(b"", False)

    @classmethod
    def get_report(cls):
        data = stats.stats()
        data.update(pid=os.getpid())
        return data


def dealer_socket( ctx: zmq.Context, address: str, identity: Optional[bytes]=None ) -> zmq.Socket:
    """ Socket for receiving incoming messages
    """
    LOGGER.debug("Connecting to %s", address)
    sock = ctx.socket(zmq.DEALER)
    sock.setsockopt(zmq.LINGER, 500)    # Needed for socket no to wait on close
    sock.setsockopt(zmq.SNDHWM, 1)      # Max 1 item on send queue
    sock.setsockopt(zmq.IMMEDIATE, 1)   # Do not queue if no peer, will block on send
    sock.setsockopt(zmq.RCVTIMEO, 1000) # Heartbeat
    sock.identity = identity or uuid.uuid1().bytes
    LOGGER.debug("Identity set to %s", sock.identity)
    sock.connect(address)
    return sock


def broadcast_socket( ctx: zmq.Context, broadcastaddr: str ) -> zmq.Socket:
    """ Socket for receiving broadcast message notifications
    """
    LOGGER.debug("Enabling broadcast notification")
    ctx = zmq.Context.instance()
    sub = ctx.socket(zmq.SUB)
    sub.setsockopt(zmq.LINGER, 500)    # Needed for socket no to wait on close
    sub.setsockopt(zmq.SUBSCRIBE, b'RESTART')
    sub.setsockopt(zmq.SUBSCRIBE, b'REPORT')
    sub.connect(broadcastaddr)
    return sub


def run_worker(address: str, handler_factory: Type[RequestHandler], 
               identity: Optional[bytes]=None, broadcastaddr: Optional[str]=None,
               maxcycles: Optional[int]=None) -> None:
    """ Enter the message loop
    """
    ctx = zmq.Context.instance()

    sock = dealer_socket( ctx, address, identity )
    if broadcastaddr:
        sub = broadcast_socket( ctx, broadcastaddr )

    # Initialize supervisor client
    supervisor = SupervisorClient()

    def get():
        client_id, corr_id, request = sock.recv_multipart()
        LOGGER.debug("RCV %s: %s", client_id, corr_id)
        return client_id, corr_id, pickle.loads(request)

    try:
        LOGGER.info("Starting ZMQ worker loop")
        completed = 0
        while maxcycles is None or (maxcycles and completed < maxcycles):
            sock.send(WORKER_READY)
            try:
                client_id, corr_id, request = get()
                supervisor.notify_busy()
                handler = handler_factory(sock, client_id, corr_id, request)
                handler.handle_message()
                completed += 1
            except zmq.error.Again:
                pass
            except zmq.ZMQError as err:
                LOGGER.error("Worker Error %d: %s", err.errno, zmq.strerror(err.errno))
            except Exception as exc:
                LOGGER.error("Worker Error %s\n%s", exc, traceback.format_exc())
                if not handler.header_written:
                    handler.status_code = 500
                    handler.send(bytes("Worker internal error".encode('ascii')))
                    # Got error 500, do not presume worker state
                    break
            finally:
                supervisor.notify_done()

            # Handle broadcast notifications
            try:
                if broadcastaddr:
                    msg = sub.recv(flags=zmq.NOBLOCK)
                    if msg==b'RESTART':
                        # There is no really way to restart
                        # so exit and let the framework restart a new worker
                        LOGGER.info("Exiting on RESTART notification")
                        break
                    elif msg==b'REPORT':
                        # Reporting asked
                        supervisor.send_report(handler_factory.get_report())
            except zmq.error.Again:
                pass
    except (KeyboardInterrupt, SystemExit):
        pass

    if broadcastaddr:
        sub.close()
    sock.close()
    LOGGER.info("Terminating Worker")


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description='Test worker')
    parser.add_argument('--host'    , metavar="host"   , default='tcp://localhost', help="router host")   
    parser.add_argument('--router'  , metavar='address', default='{host}:8881', help="router address")
    parser.add_argument('--logging' , choices=['debug', 'info', 'warning', 'error'], default='info', help="set log level")
    parser.add_argument('--identity', default="", help="Set worker identity")

    args = parser.parse_args()

    setup_log_handler(args.logging)
    print("Log level set to {}\n".format(logging.getLevelName(LOGGER.level)), file=sys.stderr)

    LOGGER.setLevel(getattr(logging, args.logging.upper()))

    run_worker(args.router.format(host=args.host), RequestHandler, 
               identity=bytes(args.identity.encode('ascii')))
    print("DONE", file=sys.stderr)




#
# Copyright 2018 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Qgis server request adapters

    Embedded qgis server in a 0MQ worker 

    see: 
        - https://qgis.org/pyqgis/master/server/QgsBufferServerResponse.html
        - https://qgis.org/pyqgis/master/server/QgsBufferServerRequest.html
"""
import os
import logging
import traceback

from typing import Dict, Optional, Any

from qgis.PyQt.QtCore import Qt, QBuffer, QIODevice, QByteArray
from qgis.core import QgsProject
from qgis.server import (QgsServerRequest,
                         QgsServerResponse,
                         QgsServerException)

from .zeromq.worker import RequestHandler, run_worker
from .qgscache.cachemanager import (get_cacheservice,
                                    preload_projects,
                                    StrictCheckingError,
                                    UnreadableResourceError,
                                    PathNotAllowedError)

from .plugins import load_plugins
from .config import confservice, configure_qgis_api, qgis_api_endpoints

LOGGER = logging.getLogger('SRVLOG')

HTTP_METHODS = {
    'GET' : QgsServerRequest.GetMethod,
    'PUT' : QgsServerRequest.PutMethod,
    'POST': QgsServerRequest.PostMethod,
    'HEAD': QgsServerRequest.HeadMethod,
    'DELETE': QgsServerRequest.DeleteMethod,
    'PATCH': QgsServerRequest.PatchMethod,
}

class Request(QgsServerRequest):

    def __init__(self, handler: RequestHandler ) -> None:
        """ Create a new QgsServerRequest from zmq handler request
        """
        req = handler.request
        
        # Recreate URL
        location = req.headers.get('X-Forwarded-Url',"")
        query    = req.query.lstrip('?')
        if query:
            location += f"?{query}"

        self._data = req.data
        super().__init__(location, HTTP_METHODS[req.method], headers=req.headers)
         
    def data(self) -> QByteArray: 
        """ Return post/put data a QByteArray
        """
        # Make sure that data is valid
        return QByteArray(self._data) if self._data else QByteArray()


class Response(QgsServerResponse):
    """ Adaptor to handler response

        The data is written at 'flush()' call.
    """

    def __init__(self, handler: RequestHandler ) -> None:
        super().__init__()
        self._handler = handler
        self._buffer = QBuffer()
        self._buffer.open(QIODevice.ReadWrite)
        self._numbytes = 0
        self._finish   = False

    def setStatusCode(self, code: int) -> None:
        if not self._handler.header_written:
            self._handler.status_code = code
        else:
            LOGGER.error("Cannot set status code after header written")

    def statusCode(self) -> int:
        return self._handler.status_code

    def finish(self) -> None:
        """ Terminate the request
        """
        self._finish = True
        self.flush()

    def flush(self) -> None:
        """ Write the data to the handler buffer 
            and flush the socket

            Headers will be written at the first call to flush()
        """
        try:
            self._buffer.seek(0)
            bytesAvail = self._buffer.bytesAvailable()
            LOGGER.debug("%s: Flushing response data: (%d bytes)",self._handler.identity, bytesAvail)
            if self._finish and bytesAvail:
                # Make sure that we have Content-length set
                self._handler.headers['Content-Length']=bytesAvail
            # Take care of the logic: if finish and not handler.header_written then there is no
            # chunk following
            send_more = not self._finish or self._handler.header_written
            if bytesAvail:
                LOGGER.debug("Sending bytes %s (send_more: %s)", bytesAvail, send_more)
                self._handler.send( bytes(self._buffer.data()), send_more )
                self._buffer.buffer().clear()
            else:
                # Return empty response
                LOGGER.debug("Sending empty response (send_more: %s)", send_more)
                self._handler.send( b'', send_more )
            # push the sentinel
            if send_more and self._finish:
                LOGGER.debug("Sending EOF")
                self._handler.send( b'', False )
        except Exception:
            LOGGER.error("Caught Exception (worker: %s, msg: %s):\n%s",
                         self._handler.identity, self._handler.msgid,
                         traceback.format_exc())
            del self._handler.headers['Content-Type']
            self.sendError(500, "Internal server error")

    def header(self, key: str) -> str:
        return self._handler.headers.get(key)

    def headers(self) -> Dict[str,str]:
        """ Return headers as dict
        """
        return self._handler.headers
        
    def io(self) -> QIODevice:
        return self._buffer

    def data(self) -> QByteArray:
        """ Return buffer data
        """
        return self._buffer.data()

    def setHeader(self, key: str, value: str) -> None:
        if not self._handler.header_written:
            self._handler.headers[key] = value
        else:
            LOGGER.error("Cannot set header after header written")

    def removeHeader(self, key: str) -> None:
        self._handler.headers.pop(key,None)

    def sendError(self, code: int, message: Optional[str]=None) -> None:
        try:
            if not self._handler.header_written:
                LOGGER.error("%s (%s)", message, code)
                self._handler.status_code = code
                self._handler.send(str(message).encode())
                self._finish = True
            else:
                LOGGER.error("Cannot set error after header written")
        except Exception:
            LOGGER.critical("Unrecoverable exception:\n%s", traceback.format_exc())


    def _clearHeaders(self) -> None:
        """ Clear headers set so far
        """
        self._handler.headers = {}
 
    def clear(self) -> None:
        self._clearHeaders()
        self.truncate()

    def headersSent(self) -> bool:
        return self._handler.header_written

    def truncate(self) -> None:
        """ Truncate buffer
        """
        self._buffer.seek(0)
        self._buffer.buffer().clear()


class QgsRequestHandler(RequestHandler):

    @classmethod
    def init_server(cls) -> None:
        if not hasattr(cls, 'qgis_server' ):
            from .utils.qgis import init_qgis_server

            # Enable qgis server verbosity
            if LOGGER.isEnabledFor(logging.DEBUG):
                os.environ['QGIS_SERVER_LOG_LEVEL'] = '0'
                os.environ['QGIS_DEBUG'] = '1'

            cache_config = confservice['projects.cache']
            if cache_config.getboolean('trust_layer_metadata'):
                os.environ['QGIS_SERVER_TRUST_LAYER_METADATA'] = 'yes'
            if cache_config.getboolean('disable_getprint'):
                os.environ['QGIS_SERVER_DISABLE_GETPRINT'] = 'yes'

            # Configure qgis api
            for name,_ in qgis_api_endpoints(enabled_only=False):
                configure_qgis_api(name)                         
            
            verbose = LOGGER.level<=logging.DEBUG or confservice.getboolean('logging','qgis_info')

            LOGGER.debug("Initializing qgis server")
            qgsserver = init_qgis_server( enable_processing=False, 
                                          logger=LOGGER, 
                                          verbose=verbose)

            serverIface = qgsserver.serverInterface()
            load_plugins(serverIface)

            if confservice['management'].getboolean('enabled'):
                from .management.apis import register_management_apis
                register_management_apis(serverIface)

            preload_projects()

            setattr(cls, 'qgis_server' , qgsserver )

    @staticmethod
    def run( router: str, identity: str="", **kwargs: Any) -> None:
        """ Run qgis server worker loop
        """
        QgsRequestHandler.init_server()

        run_worker(router, QgsRequestHandler, identity=bytes(identity.encode('ascii')), **kwargs)

    QGIS_NO_MAP_ERROR_MSG = "No project defined. For OWS services: please provide a SERVICE and a MAP parameter" 

    def handle_message(self) -> None:
        """ Override this method to handle_messages
        """
        project_location = self.request.headers.pop('X-Map-Location', None)
        ogc_scheme       = self.request.headers.pop('X-Ogc-Scheme'  , None)

        request  = Request(self)
        response = Response(self)

        if not project_location:
            # Try to get project from environment
            project_location = os.getenv("QGIS_PROJECT_FILE")

        if not project_location and ogc_scheme != 'OAF' and request.parameter('SERVICE'):
            LOGGER.error("No project defined for %s", request.parameter('SERVICE'))
            # A project is required
            exception = QgsServerException(self.QGIS_NO_MAP_ERROR_MSG, 400)
            response.write(exception)
            response.finish()
            return

        if project_location:
            iface = self.qgis_server.serverInterface()
            try:
                LOGGER.debug("Handling request: %s", self.msgid)
                project, updated = get_cacheservice().lookup(project_location)
                config_path = project.fileName()
                if updated: 
                    # Needed to cleanup cached capabilities
                    LOGGER.debug("Cleaning config cache entry %s", config_path)
                    iface.removeConfigCacheEntry(config_path)
            except StrictCheckingError:
                response.sendError(422,f"Invalid layers for project '{project_location}' - strict mode on")
            except UnreadableResourceError:
                response.sendError(422,f"Cannot read project resource '{project_location}'")
            except PathNotAllowedError:
                response.sendError(403,"Project path not allowed")
            except FileNotFoundError:
                response.sendError(404,f"Project '{project_location}' not found")
            else:
                # See https://github.com/qgis/QGIS/pull/9773
                iface.setConfigFilePath(config_path)
                self.qgis_server.handleRequest(request, response, project=project)
        else:
            self.qgis_server.handleRequest(request, response)

    @classmethod
    def get_report(cls):
        report = super(QgsRequestHandler,cls).get_report()

        def _to_json(key: str, project: QgsProject):
            return dict(
                key=key,
                filename=project.fileName(),
                last_modified=project.lastModified().toString(Qt.ISODate),
                num_layers=project.count(),
            )

        report.update(
            cache=[_to_json(k,d.project) for (k,d) in get_cacheservice().items()]
        )
        return report



def main():
    """ Run as command line interface
    """
    import os
    import sys
    import argparse
    from .version import __version__
    from .config  import (confservice, load_configuration, read_config_file, validate_config_path)
    from .logger import setup_log_handler

    parser = argparse.ArgumentParser(description='Qgis Server Worker')
    parser.add_argument('-d','--debug', action='store_true', default=False, help="debug mode") 
    parser.add_argument('-c','--config', metavar='PATH', nargs='?', dest='config',
                        default=None, help="Configuration file")
    parser.add_argument('--host'     , metavar="host"   , default='localhost' , help="router host")   
    parser.add_argument('--router'   , metavar='address', default='tcp://{host}:18080', help="router address")
    parser.add_argument('--broadcast', metavar='address', default='tcp://{host}:18090', help="broadcast address")
    parser.add_argument('--identity' , default="", help="Set worker identity")
    parser.add_argument('--rootdir'  , default=argparse.SUPPRESS, metavar='PATH', help='Path to qgis projects')
    parser.add_argument('--version'  , action='store_true', default=False, help="Return version number and exit")

    args = parser.parse_args()

    def print_version():
        program = os.path.basename(sys.argv[0])
        print("{name} {version}".format(name=program, version=__version__), file=sys.stderr)

    if args.version:
        print_version()
        sys.exit(1)

    load_configuration()

    if args.config:
        with open(args.config, mode='rt') as config_file:
            read_config_file(config_file)

    # Override config
    def set_arg( section:str, name:str ) -> None:
        if name in args:
            confservice.set( section, name, str(getattr(args,name)))

    set_arg( 'projects.cache'  , 'rootdir' )

    if args.debug:
        # Force debug mode
        confservice.set('logging', 'level', 'DEBUG')

    print_version()

    validate_config_path('projects.cache','rootdir')

    setup_log_handler(confservice.get('logging','level'))
    print("Log level set to {}\n".format(logging.getLevelName(LOGGER.level)), file=sys.stderr)

    broadcastaddr = args.broadcast.format(host=args.host)

    QgsRequestHandler.run(args.router.format(host=args.host), identity=args.identity,
                          broadcastaddr=broadcastaddr)

    print("Qgis worker terminated", file=sys.stderr)






#
# Copyright 2021 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
    Server management
"""
import os
import tornado.web
import logging

from ..logger import log_request
from ..config import confservice
from ..handlers import (BaseHandler, StatusHandler, OwsApiHandler as QgisHandler, NotFoundHandler)
from ..zeromq import client

from typing import Awaitable


LOGGER=logging.getLogger('SRVLOG')


class _PoolHandler(BaseHandler):

    def initialize(self, poolserver ) -> None:
        super().initialize()
        self._poolserver = poolserver

    def options(self) -> None:
        """ Implement OPTION for validating CORS
        """
        self.set_option_headers()


class _RestartHandler(_PoolHandler):

    def post(self, action) -> None:
        """ Handle POST method
        """
        # Broadcast 'restart' to workers
        if action == 'restart':
            # Restart workers
            self._poolserver.restart()

        self.write_json({ 'status': 'ok' }) 


class _ReportHandler(_PoolHandler):

    async def get(self) -> Awaitable[None]:
        """ Return worker reports
        """
        # Broadcast 'restart' to workers
        req = self.request
        reports = await self._poolserver.get_reports()
        for w in reports:
            for entry in w['cache']:
                entry.update(link=f"{req.protocol}://{req.host}/cache/{entry['key']}")

        self.write_json({'workers': reports, 'num_workers': self._poolserver.num_workers }) 

class _RootHandler(BaseHandler):

    def get(self) -> None:
        """ Return links to default api entries
        """
        req = self.request
        def _link( path: str, title: str ):
            return {
                'href' : f"{req.protocol}://{req.host}{path}",
                'title': title,
                'type' : "application/json",
            }

        data = dict(
            links=[
                _link("/status" , "Server status and configuration"),
                _link("/plugins", "Plugins managment"),
                _link("/cache"  , "Projects cache managment"),
                _link("/pool"   , "Workers pool status"),
            ]
        )
        self.write(data) 


    def options(self) -> None:
        """ Implement OPTION for validating CORS
        """
        self.set_option_headers()



def configure_handlers( poolserver, client: client.AsyncClient ) -> [tornado.web.RequestHandler]:
    """
    """
    kwargs =  {
        'client': client, 
        'timeout': confservice['server'].getint('timeout')
    }

    handlers = [
        (r"/"      , _RootHandler),
        (r"/status", StatusHandler),
        (r"/pool/(restart)", _RestartHandler, {'poolserver': poolserver}),
        (r"/pool/?"         ,_ReportHandler,  {'poolserver': poolserver}),
        # Forward to Qgis api handlers
        (r"/.+"             , QgisHandler, kwargs),
    ]
    return handlers


class _Management(tornado.web.Application):


    def __init__(self, poolserver, router: str) -> None:
        """
        """
        identity = bytes(f"MANAGEMENT-{os.getpid()}".encode('ascii'))
        self._broker_client = client.AsyncClient(router, identity)
        super().__init__(configure_handlers(poolserver, self._broker_client),
                         default_handler_class=NotFoundHandler)


    def log_request(self, handler: tornado.web.RequestHandler ) -> None:
        """ Write HTTP requet to the logs
        """
        log_request(handler)

    def terminate(self) -> None:
        self._broker_client.terminate()


def create_ssl_context( conf ):
    import ssl
    ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_ctx.load_cert_chain(conf['ssl_cert'],conf['ssl_key'])
    return ssl_ctx


def start_management_server( poolserver, router: str ) -> _Management:
    """ Start management server,
    """
    conf = confservice['management']

    port    = conf.getint('port')
    address = conf['interfaces']

    LOGGER.info("MANAGEMENT: running server on %s:%s",address,port)

    kwargs = {}
    if conf.getboolean('ssl'):
        LOGGER.info("MANAGEMENT: SSL enabled")
        kwargs['ssl_options'] = create_ssl_context(conf)

    app = _Management(poolserver, router)
    app.listen(port, address=address, xheaders=True, **kwargs)

    return app




#
# Copyright 2020 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from .file_handler import *     # noqa: F403,F401
from .postgres_handler import * # noqa: F403,F401

__all__=[]



#
# Copyright 2018 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Qgis server plugin managment 

"""
import sys
import logging
import configparser
import traceback

from pathlib import Path
from typing import Generator, Dict

from .config  import confservice

LOGGER = logging.getLogger('SRVLOG') 

server_plugins = {}
failed_plugins = {}


def checkQgisVersion(minver: str, maxver: str) -> bool:
    from qgis.core import Qgis

    def to_int(ver):
        major, *ver = ver.split('.')
        major = int(major)
        minor = int(ver[0]) if len(ver) > 0 else 0
        rev   = int(ver[1]) if len(ver) > 1 else 0
        if minor >= 99:
            minor = rev = 0
            major += 1
        if rev > 99:
            rev = 99
        return int("{:d}{:02d}{:02d}".format(major,minor,rev))


    version = to_int(Qgis.QGIS_VERSION.split('-')[0])
    minver  = to_int(minver) if minver else version
    maxver  = to_int(maxver) if maxver else version

    return minver <= version <= maxver

    

def find_plugins(path: str) -> Generator[str,None,None]:
    """ return list of plugins in given path
    """
    path = Path(path)
    for plugin in path.glob("*"):
        LOGGER.debug("Looking for plugin in %s", plugin)
        if not plugin.is_dir():
            continue

        metadatafile = plugin / 'metadata.txt'
        if not metadatafile.exists():
            continue

        if not (plugin / '__init__.py').exists():
            LOGGER.warning("Found metadata file but no entry point !")
            continue

        cp = configparser.ConfigParser()

        try:
            with metadatafile.open(mode='rt') as f:
                cp.read_file(f)

            if not cp['general'].getboolean('server'):
                LOGGER.warning("%s is not a server plugin", plugin)
                continue

            minver = cp['general'].get('qgisMinimumVersion')
            maxver = cp['general'].get('qgisMaximumVersion')

        except Exception as exc:
            LOGGER.error("Error reading plugin metadata '%s': %s",metadatafile,exc)
            continue

        if not checkQgisVersion(minver,maxver):
            LOGGER.warning("Unsupported version for %s. Discarding", plugin)
            continue

        yield plugin.name



def load_plugins(serverIface: 'QgsServerInterface') -> bool: # noqa F821
    """ Start all plugins

        Return False if some plugins can no be loaded
    """

    plugin_path = confservice.get('server','pluginpath')
    if not plugin_path:
        return

    LOGGER.info("Initializing plugins from %s", plugin_path)
    sys.path.append(plugin_path)

    for plugin in find_plugins(plugin_path):
        try:
            __import__(plugin)

            package = sys.modules[plugin] 

            # Initialize the plugin
            server_plugins[plugin] = package.serverClassFactory(serverIface)
            LOGGER.info("Loaded plugin %s",plugin)
        except Exception:
            strace = traceback.format_exc()
            LOGGER.error("Error loading plugin '%s'\n%s", plugin, strace)
            failed_plugins[plugin] = strace
    

def plugin_metadata( plugin: str ) -> Dict:
    """ Return plugin metadata
    """
    if plugin not in server_plugins:
        return

    # Read metadata
    path = Path(sys.modules[plugin].__file__)
    metadatafile = path.parent / 'metadata.txt'
    if not metadatafile.exists():
        return

    with metadatafile.open(mode='rt') as f:
        cp = configparser.ConfigParser()
        cp.read_file(f)
        metadata = { s: dict(p.items()) for s,p in cp.items() }
        metadata.pop('DEFAULT',None)
        metadata.update(path=str(path))
        return metadata


def plugin_list():
    """ Iterate over loaded plugins
    """
    return (k for k in server_plugins.keys())




#
# Copyright 2018 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import asyncio
import tornado.platform.asyncio
from tornado.testing import AsyncHTTPTestCase
import lxml.etree
import logging
import json

from typing import Any, Dict, Optional

from .config import load_configuration
from .runtime import (Application, 
                      create_poolserver,
                      create_broker_process,
                      configure_ipc_addresses)

LOGGER = logging.getLogger('SRVLOG')

NAMESPACES = {
    'xlink': "http://www.w3.org/1999/xlink",
    'wms': "http://www.opengis.net/wms",
    'wfs': "http://www.opengis.net/wfs",
    'wcs': "http://www.opengis.net/wcs",
    'wps': "http://www.opengis.net/wps/1.0.0",
    'ows': "http://www.opengis.net/ows/1.1",
    'gml': "http://www.opengis.net/gml",
    'xsi': "http://www.w3.org/2001/XMLSchema-instance"
}

class TestRuntime:

    def __init__(self) -> None:
        self.started = False
           
    def start(self) -> None:
        if self.started:
            return

        workers = 1
        load_configuration()        
        self.ipcaddr = configure_ipc_addresses(workers)
        self._broker = create_broker_process(self.ipcaddr)
        self._pool   = create_poolserver(workers)
        self.started = True

    def stop(self) -> None:
        if not self.started:
            return
        self._pool.terminate()
        self._broker.terminate()
        self._broker.join()

    @classmethod
    def instance(cls) -> 'TestRuntime':
        if not hasattr(cls,'_instance'):
            cls._instance = TestRuntime()
        return cls._instance


class HTTPTestCase(AsyncHTTPTestCase):

    def setUp(self) -> None:
        super().setUp()
        self.logger   = LOGGER
        self.client   = OWSTestClient(self)

    def tearDown(self) -> None:
        self._application.terminate()
        super().tearDown()

    def get_app(self) -> Application:
        ipcaddr = TestRuntime.instance().ipcaddr
        self._application = Application(ipcaddr)
        return self._application
    
    def get_new_ioloop(self) -> tornado.platform.asyncio.AsyncIOLoop:
        """
        Needed to make sure that I can also run asyncio based callbacks in our tests
        """
        # Create a new IO loop et set it as default
        io_loop = tornado.platform.asyncio.AsyncIOLoop()
        asyncio.set_event_loop(io_loop.asyncio_loop)
        return io_loop


class HTTPTestResponse:

    def __init__(self, http_response) -> None:
        self.http_response = http_response
        if self.headers.get('Content-Type','').find('text/xml')==0:
            self.xml = lxml.etree.fromstring(self.content)

    @property
    def content(self) -> Any:
        return self.http_response.body

    @property
    def status_code(self) -> int:
        return self.http_response.code

    @property
    def headers(self):
        return self.http_response.headers

    def xpath(self, path: str) -> 'xpath':
        return self.xml.xpath(path, namespaces=NAMESPACES)

    def xpath_text(self, path: str) -> str:
        return ' '.join(e.text for e in self.xpath(path))

    def json(self) -> Any:
        return json.loads(self.content)


class OWSTestClient:

    def __init__(self, testcase: HTTPTestCase) -> None:
        self._testcase = testcase

    def post(self, data: Any, headers: Optional[Dict]=None, path: str='/ows/') -> HTTPTestResponse:
        return HTTPTestResponse(self._testcase.fetch(path, method='POST', body=data, raise_error=False,
                                headers=headers))

    def get(self, query: str, headers: Optional[Dict]=None, path: str='/ows/') -> HTTPTestResponse:
        return HTTPTestResponse(self._testcase.fetch(path + query, raise_error=False, 
                                headers=headers))

    def put(self, data: Any, headers: Optional[Dict]=None, path:  str='/ows/') -> HTTPTestResponse:
        return HTTPTestResponse(self._testcase.fetch(path, method='PUT', body=data, raise_error=False,
                                headers=headers))

    def post_xml(self, doc: lxml.etree.Element, headers: Optional[Dict]=None, path: str='/ows/') -> HTTPTestResponse:
        return self.post(data=lxml.etree.tostring(doc, pretty_print=True), 
                         headers=headers, path=path)

    def options( self, headers: Optional[Dict]=None, path: str='/ows/') -> HTTPTestResponse:
        return HTTPTestResponse(self._testcase.fetch(path, method='OPTIONS', 
                                headers=headers, raise_error=False))




#
# Copyright 2018 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from collections import namedtuple

WORKER_READY=b"ready"

# Message structure

RequestMessage = namedtuple( "RequestMessage", (
    "query",
    "headers",
    "method",
    "data"
))


ReplyMessage = namedtuple( "ReplyMessage", (
    "status",
    "headers",
    "data",
))





#
# Copyright 2018 3liz
# Author: David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Base Request handler 
"""
import tornado.web
import logging
import json
from tornado.web import HTTPError # noqa F401

from urllib.parse import urlencode

from typing import Any, Union, Optional

from ..version import __version__
from ..config import confservice

LOGGER = logging.getLogger('SRVLOG')


class BaseHandler(tornado.web.RequestHandler):
    """ Base class for HTTP request hanlers
    """
    def initialize(self) -> None:
        super().initialize()
        self._links    = []
        self.connection_closed = False
        self.logger = LOGGER
        self._cfg   = confservice['server']
        self._cross_origin = self._cfg.getboolean('cross_origin')

    def prepare(self) -> None:
        self.has_body_arguments = len(self.request.body_arguments)>0
        # Replace query arguments to upper case:
        self.request.arguments = { k.upper():v for (k,v) in self.request.arguments.items() }

    def encode_arguments(self) -> str:
        return '?'+urlencode({k:v[0] for k,v in self.request.arguments.items()})

    def compute_etag(self) -> None:
        # Disable etag computation
        pass

    def set_default_headers(self) -> None:
        """ Override defaults HTTP headers 
        """
        self.set_header("Server",__version__)

    def on_connection_close(self) -> None:
        """ Override, log and set 'connection_closed' to True
        """
        self.connection_closed = True
        self.logger.warning("Connection closed by client: {}".format(self.request.uri))

    def set_option_headers(self, allow_header: Optional[str]=None) -> None:
        """  Set correct headers for 'OPTION' method
        """
        if not allow_header:
            allow_header = ', '.join( me for me in self.SUPPORTED_METHODS if hasattr(self, me.lower()) )
        
        self.set_header("Allow", allow_header)
        if self.set_access_control_headers():
            # Required in CORS context
            # see https://developer.mozilla.org/fr/docs/Web/HTTP/M%C3%A9thode/OPTIONS
            self.set_header('Access-Control-Allow-Methods', allow_header)

    def set_access_control_headers(self) -> bool:
        """  Handle Access control and cross origin headers (CORS)
        """
        origin = self.request.headers.get('Origin')
        if origin:
            if self._cross_origin:
                self.set_header('Access-Control-Allow-Origin', '*')
            else:
                self.set_header('Access-Control-Allow-Origin', origin)
                self.set_header('Vary', 'Origin')
            return True
        else:
            return False

    def write_json(self, chunk: Union[str,dict]) -> None:
        """ Write body as json

            The method will also set CORS implicitely for any origin
            If this a security issue, we should allow it
            explicitely. 
        """
        if isinstance(chunk, dict):
            chunk = json.dumps(chunk, sort_keys=True)
        self.set_header('Content-Type', 'application/json;charset=utf-8')
        self.set_access_control_headers()
        self.write(chunk)

    def write_error(self, status_code: int, **kwargs: Any) -> None:
        """ Override, format error as json
        """
        message = self._reason

        if "exc_info" in kwargs:
            exception = kwargs['exc_info'][1]
            # Error was caused by a exception
            message = "{}".format(exception)
               
        self.logger.error("%s", message)
        response = dict(status="error" if status_code != 200 else "ok",
                        httpcode = status_code,
                        error    = { "message": message })

        self.write_json(response)
        self.finish()

    def proxy_url(self, http_proxy: bool, endpoint: Optional[str]=None) -> str:
        """ Return the proxy_url
        """
        # Replace the status url with the proxy_url if any
        req = self.request
        if http_proxy:
            proxy_url = self._cfg.get('proxy_url') or \
                req.headers.get('X-Forwarded-Url') or \
                f"{req.protocol}://{req.host}/"
            if endpoint:
                proxy_url = f"{proxy_url.rstrip('/')}{endpoint}"
        else:
            # No proxy to handle: return the full path
            proxy_url = f"{req.protocol}://{req.host}{req.path}"
        return proxy_url



class NotFoundHandler(BaseHandler):
    def prepare(self):  # for all methods
        raise HTTPError(
            status_code=404,
            reason="Invalid resource path."
        )





#
# Author David Marteau
#
# This code is derived from the tornado implementation and is released under the same
# applicable License
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#
#
""" Handle fork

    The code is derived from the tornado implementation:
    see http://www.tornadoweb.org/en/stable/_modules/tornado/process.html#fork_processes
"""
import sys
import os
import logging
import signal
import errno

from tornado.util import errno_from_exception
from typing import Union


_task_id  = None
_ppid     = None
_children = {}

def _start_child(i: int) -> Union[int,None]:

    global _children

    pid = os.fork()
    if pid == 0:
        # child process
        global _task_id
        _task_id = i
        return i
    else:
        _children[pid] = i
        return None


def fork_processes(num_processes: int) -> Union[int,None]:
    """Starts multiple worker processes.

    In each child process, ``fork_processes`` returns its *task id*, a
    number between 0 and ``num_processes``.  
    
    """
    global _ppid
    assert _ppid is None

    global _task_id
    assert _task_id is None

    global _children

    _ppid = os.getpid()    

    for i in range(num_processes):
        id = _start_child(i)
        if id is not None:
            _children.clear()
            return id

    return None


def manage_processes( max_restarts: int, logger: logging.Logger = None ) -> Union[int, None]:
    """ Manage child processes 

    Processes that exit
    abnormally (due to a signal or non-zero exit status) are restarted
    with the same id (up to ``max_restarts`` times).  In the parent
    process, ``fork_processes`` returns None if all child processes
    have exited normally, but will otherwise only exit by throwing an
    exception.
    """

    assert _ppid == os.getpid(), "can only manage processes created in the current process"

    num_restarts = 0

    logger = logger or logger.getLogger()

    global _children

    while _children:
        try:
            pid, status = os.wait()
        except OSError as e:
            if errno_from_exception(e) == errno.EINTR:
                continue
            raise
        if pid not in _children:
            continue
        id = _children.pop(pid)
        if os.WIFSIGNALED(status):
            logger.warning("child %d (pid %d) killed by signal %d, restarting",
                           id, pid, os.WTERMSIG(status))
        elif os.WEXITSTATUS(status) != 0:
            logger.warning("child %d (pid %d) exited with status %d, restarting",
                           id, pid, os.WEXITSTATUS(status))
        else:
            logger.debug("child %d (pid %d) exited normally", id, pid)
            continue
        num_restarts += 1
        if num_restarts > max_restarts:
            raise RuntimeError("Too many child restarts, giving up")
        new_id = _start_child(id)
        if new_id is not None:
            return new_id
    # All child processes exited cleanly, so exit the master process
    # instead of just returning to right after the call to
    # fork_processes (which will probably just start up another IOLoop
    # unless the caller checks the return value).
    sys.exit(0)


def task_id() -> int:
    """Returns the current task id, if any.

    Returns None if this process was not created by `fork_processes`.
    """
    global _task_id
    return _task_id


def terminate_childs() -> None:
    """ Terminate all childs
    """
    if _ppid is None: 
        # Nothing has bee started so far
        return

    assert _ppid == os.getpid(), "can only terminate processes created in the current process"
    global _children
    for pid in _children:
        os.kill(pid, signal.SIGTERM)






""" Watch file change
"""
import os
import functools
import logging
import traceback

from typing import Callable, Mapping, Union, List
from tornado import ioloop

LOGGER = logging.getLogger('SRVLOG')

UpdateFunc = Callable[[List[str]],None]


def watchfiles(watched_files: List[str], updatefunc: UpdateFunc,  check_time: int=500) -> ioloop.PeriodicCallback:
    """Begins watching source files for changes.
    """
    modify_times = {}
    callback = functools.partial(_update_callback, updatefunc, watched_files, modify_times)
    scheduler = ioloop.PeriodicCallback(callback, check_time)
    return scheduler


def _update_callback( updatefunc: UpdateFunc, watched_files: List[str], modify_times: Mapping[float,str]) -> None:
    """ Call update funcs when modified files
    """
    modified_files = [path for path in watched_files if _check_file(modify_times, path) is not None]
    if len(modified_files) > 0:
        LOGGER.debug("running update hook for %s", modified_files)
        updatefunc( modified_files )


def _check_file(modify_times: Mapping[float,str], path: str) -> Union[str,None]:
    try:
        modified = os.stat(path).st_mtime
    except FileNotFoundError:
        # Do not care if file do not exists
        return
    except Exception:
        traceback.print_exc()
        LOGGER.error("Error while checking file %s")
        return
    if path not in modify_times:
        modify_times[path] = modified
        return
    if modify_times[path] != modified:
        modify_times[path] = modified
        return path




#
# Copyright 2018 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Zmq asynchrone client
"""

import sys
import asyncio
import zmq
import zmq.asyncio
import pickle
import logging
import uuid
import traceback

from typing import Mapping, Any, Tuple

from .messages import RequestMessage

from ..logger import setup_log_handler

LOGGER=logging.getLogger('SRVLOG')

class RequestTimeoutError(Exception):
    pass

class RequestGatewayError(Exception):
    pass

class RequestProxyError(Exception):
    pass


class AsyncResponseHandler:
    def __init__(self, correlation_id: bytes) -> None:

        loop = asyncio.get_event_loop()

        self.correlation_id = correlation_id
        self.headers = None
        # Create a future for sending the result
        self._future = loop.create_future()
        self._chunks = None
        self._has_more = False
        self.data = None
    
    def _set_exception(self, exc: Exception) -> None:
        self._has_more = False
        self._future.set_exception(exc)

    def _set_result(self, data: bytes) -> None:
        """ Set raw results from request

            This method send the result to the stored
            future.
        """
        if self.headers is None:
            status, hdrs, body = pickle.loads(data)
            self.headers = hdrs
            self.data    = body
            self.status  = status
            # We are waiting for more data
            # Create a queue to collect the remaining chunks
            if self.status == 206:
                self._has_more = True
                self._chunks = asyncio.Queue()
            # Send the result
            self._future.set_result(self)
        elif self._has_more:
            body, has_more = pickle.loads(data)
            self._has_more = has_more
            self._chunks.put_nowait((body,has_more))

    def _done(self) -> bool:
        """ Check if there is more data to come
        """
        return not self._has_more and self._future.done()

    async def _get(self, timeout: int) -> Any:
        """ wait for result and return the parsed 
            result
        """
        try:
            return await asyncio.wait_for(self._future, timeout)
        except asyncio.TimeoutError:
            self._has_more = False
            raise RequestTimeoutError()

    async def _next_chunk(self, timeout: int) -> Tuple[bytes,bool]:
        """ Get next chunk
        """
        try:
            return await asyncio.wait_for(self._chunks.get(), timeout)
        except asyncio.TimeoutError:
            self._has_more = False
            raise RequestTimeoutError()


class AsyncClient:
    """ Async DEALER ZMQ client
    """
    def __init__(self, address: str , identity: bytes=None) -> None:
        
        context = zmq.asyncio.Context.instance()

        self.identity = identity or uuid.uuid1().bytes

        sock = context.socket(zmq.DEALER)
        sock.setsockopt(zmq.LINGER, 500)    # Needed for socket no to wait on close
        sock.setsockopt(zmq.IMMEDIATE, 1)   # Do not queue if there is no connection
        sock.identity = self.identity
        sock.connect(address)

        self._running = False
        self._handlers = {}
        self._socket   = sock
        self._polling  = False
        LOGGER.info("Starting client %s", self.identity)

    async def _poll(self) -> None:
        """ Handle incoming messages
        """
        self._polling = True
        while self._handlers:
            try:
                correlation_id, data, *rest = await self._socket.recv_multipart()
                # Get if there is a future pending for that message
                try:
                    handler = self._handlers[correlation_id]
                    if rest and data == b'ERR':
                        handler._set_exception(RequestProxyError(rest[0]))
                    else:
                        handler._set_result(data)
                    # Remove handlers from the heap if we are done
                    if handler._done():
                        self._handlers.pop(correlation_id,None)
                except KeyError:
                    LOGGER.warning("%s: No pending future found for message %s",self.identity, correlation_id)
            except zmq.ZMQError as err:
                LOGGER.error("%s error:  zmq error: %s (%s)", self.identity, zmq.strerror(err.errno),err.errno)
            except Exception as err:
                LOGGER.error("%s exception %s\n%s", self.identity, err, traceback.format_exc())
        self._polling = False

    async def fetch( self, query: str, method: str='GET', headers: Mapping[str,str]={}, 
                     data: Any=None, timeout: int=5) -> Any:
        """ Send a request message to the worker
        """
        # Send request
        request = pickle.dumps(RequestMessage(query,headers=headers,method=method,data=data),-1)
        correlation_id = uuid.uuid1().bytes
        assert correlation_id not in self._handlers
        try:
            await self._socket.send_multipart([correlation_id, request], flags=zmq.DONTWAIT)
        except zmq.ZMQError as err:
            LOGGER.error("%s (%s)", zmq.strerror(err.errno), err.errno)
            raise RequestGatewayError()
        
        # Create response handler and register it
        handler = AsyncResponseHandler(correlation_id)
        self._handlers[correlation_id] = handler
        # Run poller if needed
        if not self._polling:
            asyncio.ensure_future(self._poll())

        # Wait for response
        try:
            return await handler._get(timeout)
        except Exception:
            self._handlers.pop(correlation_id,None)
            raise

    async def fetch_more( self, response, timeout=5 ):
        """ Request next chunk
        """
        try:
            while True:
                data, has_more = await response._next_chunk(timeout) 
                if not has_more:
                    break
                yield  data
        except Exception:
            self._handlers.pop(response.correlation_id,None)
            raise

    def terminate(self) -> None:
        LOGGER.info("Terminating client %s", self.identity)
        self._running = False
        self._futures = {}
        self._socket.close()
    

if __name__ == '__main__':
    import signal
    import argparse
    from time import sleep

    parser = argparse.ArgumentParser(description='Test Client')
    parser.add_argument('--host'      , metavar="host"   , default='tcp://localhost', help="router host")
    parser.add_argument('--router'    , metavar='address', default='{host}:8880', help="router address")
    parser.add_argument('--logging'   , choices=['debug', 'info', 'warning', 'error'], default='info', help="set log level")
    parser.add_argument('--identity'  , default='', help="Set worker identity")
    parser.add_argument('--count'     , default=1, type=int, help="Number of requests")

    args = parser.parse_args()

    setup_log_handler(args.logging)
    print("Log level set to {}\n".format(logging.getLevelName(LOGGER.level)), file=sys.stderr)

    LOGGER.setLevel(getattr(logging, args.logging.upper()))

    client = AsyncClient(args.router.format(host=args.host),bytes(args.identity.encode('ascii')))
    sleep(1) # Give some time to connection to establish

    async def fetch(index):
        try:
            response = await client.fetch(query="?service=WMS", data=b"Hello world from %d" % index)
            print("%d -> response = %s" % (index,response.data))
            async for chunk in client.fetch_more(response):
                print("%d -> chunk = %s" % (index,chunk))
        except RequestTimeoutError:
            LOGGER.error("%d -> TIMEOUT", index)
        except RequestGatewayError:
            LOGGER.error("%d -> GATEWAY ERROR", index)

    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, loop.stop)
    loop.run_until_complete(asyncio.wait([fetch(i+1) for i in range(args.count)]))
    
    client.terminate()

    print("DONE", file=sys.stderr)




#
# Copyright 2018 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import sys
import asyncio
import logging
import signal

import tornado.web
import tornado.platform.asyncio

from multiprocessing import Process

from typing import Optional, Mapping, List

from .logger import log_request
from .config import confservice, qgis_api_endpoints

from .handlers import (StatusHandler, 
                       OwsHandler, 
                       OwsFilterHandler, 
                       OwsApiHandler,
                       OwsApiFilterHandler,
                       PingHandler, 
                       NotFoundHandler)

from .zeromq import client, broker

from .utils import process
from .qgspool import create_poolserver

from .monitor import Monitor
from .stats import Stats

from pyqgisservercontrib.core.filters import ServerFilter

LOGGER=logging.getLogger('SRVLOG')


def load_access_policies() -> Optional[Mapping[str,List[ServerFilter]]]:
    """ Create filter list
    """
    if not confservice.getboolean('server','enable_filters'):
        return None

    confservice.set('server','access_policy_version','2')

    import pyqgisservercontrib.core.componentmanager as cm

    collection = []
    cm.register_entrypoints('qgssrv_contrib_access_policy', collection) 

    if not collection:
        return None

    # Retrieve filters
    filters = { "": [] }
    for filt in collection:
        uri = filt.uri or ""
        fls = filters.get(uri,[])
        fls.append(filt)
        filters[uri] = fls
    # Sort filters
    for flist in filters.values():
        flist.sort(key=lambda f: f.pri, reverse=True)
    return filters


def configure_handlers( client: client.AsyncClient ) -> [tornado.web.RequestHandler]:
    """ Configure request handlers
    """
    cfg = confservice['server']

    monitor = Monitor.initialize()

    root = r"/ows"

    ows_kwargs = {
        'client'      : client,
        'monitor'     : monitor,
        'timeout'     : cfg.getint('timeout'),
        'http_proxy'  : cfg.getboolean('http_proxy'),
        'allowed_hdrs': tuple(k.upper() for k in cfg.get('allow_headers').split(','))
    }

    end = r"(?:\.html|\.json|/?)"

    ows_api_endpoints = [
        rf"/wfs3{end}",
        rf"/wfs3/collections(?:/[^/]+(?:/items)?)?{end}",
        rf"/wfs3/conformance{end}",
        rf"/wfs3/api{end}",
        r"/wfs3/static/.*",
    ] 

    handlers = [
        (r"/ping", PingHandler),
    ]

    def add_handler( path, handler, kwargs ):
        LOGGER.debug("*** Adding handler for: %s", path)        
        handlers.append( (path, handler, kwargs) )

    # Server status page
    if cfg.getboolean('status_page'):
        handlers.append( ("/status/?", StatusHandler) )

    # Load filters
    filters = load_access_policies()
    if filters:
        for uri,fltrs in filters.items():
            kw = ows_kwargs.copy()
            kw.update( filters = fltrs)
            # Add ow endpoint
            path = f"{root}/{uri.strip('/')}" if uri else root
            # Add service endpoint
            add_handler( f"{path}(?P<endpoint>/?)", OwsFilterHandler, kw )
            for endp in ows_api_endpoints:
                add_handler( rf"{path}(?P<endpoint>{endp})", OwsApiFilterHandler, kw )
    else:
        add_handler( rf"{root}(?P<endpoint>/?)", OwsHandler, ows_kwargs)
        for endp in ows_api_endpoints:
            add_handler( rf"{root}(?P<endpoint>{endp})", OwsApiHandler, ows_kwargs )

    #
    # Add qgis api endpoints
    #
    for name,endpoint in qgis_api_endpoints():
        add_handler( rf"(?P<endpoint>/{endpoint.strip('/')}/.*)", OwsApiHandler, ows_kwargs )

    return handlers


class Application(tornado.web.Application):

    def __init__(self, router: str) -> None:
        """
        """
        identity = confservice['zmq']['identity']
        identity = "{}-{}".format(identity,os.getpid())

        self._broker_client = client.AsyncClient(router, bytes(identity.encode('ascii')))
        self.stats = Stats()

        super().__init__(configure_handlers(self._broker_client),
                         default_handler_class=NotFoundHandler)

    def log_request(self, handler: tornado.web.RequestHandler ) -> None:
        """ Write HTTP requet to the logs
        """
        log_request(handler)        

    def terminate(self) -> None:
        self._broker_client.terminate()


def terminate_handler( signum: int, frame ) -> None:
    if signum == signal.SIGTERM:
        if process.task_id() is None:
            sys.stderr.write("Terminating child processes.\n")
            process.terminate_childs()


def set_signal_handlers() -> None:
    signal.signal(signal.SIGTERM, terminate_handler)
    signal.signal(signal.SIGINT , terminate_handler)


def setuid( username: str) -> None:
    """ setuid to username uid """
    from pwd import getpwnam, getpwuid
    pw = getpwnam(username)
    os.setgid(pw.pw_gid)
    os.setuid(pw.pw_uid)
    LOGGER.info("Setuid to user {} ({}:{})".format(getpwuid(os.getuid()).pw_name, os.getuid(), os.getgid()))


def create_broker_process( ipcaddr: str ) -> Process:
    """ Create a brker process
    """
    cfg = confservice['zmq']

    LOGGER.info("Starting broker process")
    p = Process(target=broker.run_broker, kwargs=dict(
                inaddr   = ipcaddr,
                outaddr  = cfg['bindaddr'],
                maxqueue = cfg.getint('maxqueue'),
                timeout  = cfg.getint('timeout')))
    p.start()
    return p


def configure_ipc_addresses(workers: int) -> str:
    """ Create ipc socket path
    """
    pid = os.getpid()
    ipc_path = f"/tmp/qgssrv/broker/{pid}/"
    os.makedirs(os.path.dirname(ipc_path), exist_ok=True)
    ipcaddr = f"ipc://{ipc_path}0"
    # Use ipc sockets for managed workers
    confservice.set('zmq','ipcpath', ipc_path)
    if workers > 0:
        confservice.set('zmq','bindaddr'     , f"ipc://{ipc_path}pool0")
        confservice.set('zmq','broadcastaddr', f"ipc://{ipc_path}broadcast0")

    return ipcaddr


def create_ssl_options():
    """ Create an ssl context
    """
    import ssl
    cfg = confservice['server']
    ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_ctx.load_cert_chain(cfg['ssl_cert'],cfg['ssl_key'])
    return ssl_ctx


def run_server( port: int, address: str="", jobs: int=1,  user: str=None, workers: int=0) -> None:
    """ Run the server

        :param port: port number
        :param address: interface address to bind to (default to any)
        :param jobs: Number of jobs to fork (default to 1: i.e no fork)
        :user: User to setuid after opening ports (default no setuid)
    """
    import traceback
    from tornado.netutil import bind_sockets
    from tornado.httpserver import HTTPServer

    kwargs = {}

    ipcaddr = configure_ipc_addresses(workers)

    application = None
    server = None

    if user:
        setuid(user)

    worker_pool   = None
    broker_pr     = None

    # Setup ssl config
    if confservice.getboolean('server','ssl'):
        LOGGER.info("SSL enabled")
        kwargs['ssl_options'] = create_ssl_options()

    if confservice.getboolean('server','http_proxy'):
        LOGGER.info("Proxy configuration enabled")
        kwargs['xheaders'] = True

    # Run
    try:
        # Fork processes
        # This is a *DEPRECATED* feature
        if jobs > 1:

            def close_sockets(sockets):
                for sock in sockets:
                    sock.close()

            print(("DEPRECATION WARNING: "
                   "the 'jobs' option is deprecated "
                   "and will be removed in near future"),
                  file=sys.stderr, flush=True)

            sockets = bind_sockets(port, address=address)
            if  process.fork_processes(jobs) is None: # We are in the main process
                close_sockets(sockets)
                broker_pr   = create_broker_process(ipcaddr)
                worker_pool = create_poolserver(workers) if workers>0 else None
                set_signal_handlers()

                # Note that manage_processes(...) never return in main process 
                # and call exit(0) which will be caught by SystemExit exception
                process.manage_processes(max_restarts=5, logger=LOGGER)

                assert False, "Not Reached"
        else:
            broker_pr   = create_broker_process(ipcaddr)
            worker_pool = create_poolserver(workers) if workers>0 else None
            sockets = bind_sockets(port, address=address)

        LOGGER.info("Running server on port %s:%s", address, port)

        application = Application(ipcaddr)

        # Init HTTP server
        server = HTTPServer(application, **kwargs)
        server.add_sockets(sockets)

        management = None
        # Activate management
        if confservice['management'].getboolean('enabled'):
            from .management.server import start_management_server
            management = start_management_server(worker_pool,ipcaddr)
            management.stats = application.stats

        # Initialize pool supervisor
        worker_pool.start_supervisor()

        loop = asyncio.get_event_loop()
        loop.add_signal_handler(signal.SIGTERM, lambda: loop.stop())
        LOGGER.info("Starting processing requests")
        loop.run_forever()
    except Exception:
        traceback.print_exc()
        if process.task_id() is not None:
            # Let a chance to the child process to 
            # restart
            raise
        else: 
            # Make sure that child processes are terminated
            print("Terminating child processes", file=sys.stderr)
            process.terminate_childs()
    except KeyboardInterrupt:
        print("Keyboard Interrupt", file=sys.stderr)
    except SystemExit as exc:
        print(exc, file=sys.stderr )

    # Teardown
    if server is not None:
        server.stop()
    if management is not None:
        management.terminate()
        management = None
    if application is not None:
        application.terminate()
        application = None
        print("{}: Server instance stopped".format(os.getpid()), file=sys.stderr)
    if process.task_id() is None:
        if worker_pool:
            print("Stopping workers")
            worker_pool.terminate()
        if broker_pr:
            print("Stopping broker")
            broker_pr.terminate()
            broker_pr.join()


    print("Server shutdown", file=sys.stderr)



#
# Copyright 2018 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import sys
import logging
import argparse

from typing import List

from .version import __description__, __manifest__
from .logger import setup_log_handler
from .config import (confservice, read_config_file,
                     validate_config_path, load_configuration)

from .utils.qgis import print_qgis_version
from .runtime import run_server

LOGGER = logging.getLogger('SRVLOG')



def print_version(verbose=False) -> None:
    """ Display version infos
    """
    program = os.path.basename(sys.argv[0])
    print("{program} {version} (build {buildid},commit {commitid})".format(program=program,**__manifest__),
          file=sys.stderr)
    print_qgis_version(verbose=verbose)


def read_configuration(argv: List[str]=None) -> argparse.Namespace:
    """ Parse command line and read configuration file
    """
    if argv is None:
        argv = sys.argv[1:]

    cli_parser = argparse.ArgumentParser(description=__description__)

    cli_parser.add_argument('-d','--debug', action='store_true', default=False, help="debug mode") 
    cli_parser.add_argument('-c','--config', metavar='PATH', nargs='?', dest='config',
                            default=None, help="Configuration file")
    cli_parser.add_argument('--version', action='store_true', 
                            default=False, help="Return version infos and exit")
    cli_parser.add_argument('-p','--port'    , type=int, help="http port", dest='port', default=argparse.SUPPRESS)
    cli_parser.add_argument('-b','--bind'    , metavar='IP',  default=argparse.SUPPRESS, help="interface to bind to", dest='interfaces')
    cli_parser.add_argument('-w','--workers' , metavar='NUM', type=int, default=argparse.SUPPRESS, help="num workers", dest='workers')
    cli_parser.add_argument('-j','--jobs'    , metavar='NUM', type=int, default=1, help="[DEPRECATED] num server instances", dest='jobs')
    cli_parser.add_argument('-u','--setuid'  , default='', help="uid to switch to", dest='setuid')
    cli_parser.add_argument('--rootdir'  , default=argparse.SUPPRESS, metavar='PATH', help='path to qgis projects')
    cli_parser.add_argument('--proxy'    , action='store_true', default=False, help='run only as proxy')

    args = cli_parser.parse_args(argv)

    if args.version:
        print_version(verbose=args.debug)
        sys.exit(1)
    else:
        print_version()

    load_configuration()

    if args.config:
        with open(args.config, mode='rt') as config_file:
            read_config_file(config_file)

    # Override config
    def set_arg( section:str, name:str ) -> None:
        if name in args:
            confservice.set( section, name, str(getattr(args,name)))

    set_arg( 'projects.cache'  , 'rootdir' )
    set_arg( 'server' , 'interfaces')
    set_arg( 'server' , 'port' )
    set_arg( 'server' , 'workers' )

    if args.debug:
        # Force debug mode
        confservice.set('logging', 'level', 'DEBUG')

    # set log level
    setup_log_handler(confservice.get('logging','level'))
    print("Log level set to {}\n".format(logging.getLevelName(LOGGER.level)), file=sys.stderr)

    conf = confservice['server']
    args.port       = conf.getint('port')
    args.workers    = conf.getint('workers')
    args.interfaces = conf.get('interfaces')

    return args


def main() -> None:
    """ Run the server as cli command
    """
    args = read_configuration()

    workers = args.workers
    if not args.proxy:
        validate_config_path('projects.cache','rootdir')
    else:
        # Do not run any qgis workers
        workers = 0

    run_server( port=args.port, address=args.interfaces, jobs=args.jobs, user=args.setuid, workers=workers )

    




#
# Copyright 2020 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" File protocol handler
"""
import logging
import urllib.parse

from typing import Tuple, Optional
from datetime import datetime
from pathlib import Path

from qgis.core import QgsProject
from pyqgisservercontrib.core import componentmanager

LOGGER = logging.getLogger('SRVLOG')

ALLOWED_SFX=('.qgs','.qgz')

__all__= []

@componentmanager.register_factory('@3liz.org/cache/protocol-handler;1?scheme=file')
class FileProtocolHandler:
    """ Handle file protocol
    """

    def __init__(self):
        pass

    def get_project( self, url: urllib.parse.ParseResult, strict: Optional[bool]=None,
                     project: Optional[QgsProject]=None,
                     timestamp: Optional[datetime]=None) -> Tuple[QgsProject, datetime]:
        """ Create or return a proect
        """
        # Securit check
        path = Path(url.path)   
        if not path.is_absolute():
            raise ValueError("file path must be absolute not %s" % path)
    
        exists = False
        if path.suffix not in ALLOWED_SFX:
            for sfx in ALLOWED_SFX:
                path = path.with_suffix(sfx)
                exists = path.is_file()
                if exists:
                    break
        else:
            exists = path.is_file()

        if not exists:
            LOGGER.error("File protocol handler: File not found: %s", str(path)) 
            raise FileNotFoundError(str(path))

        modified_time = datetime.fromtimestamp(path.stat().st_mtime)
        if timestamp is None or timestamp < modified_time:
            cachmngr  = componentmanager.get_service('@3liz.org/cache-manager;1')
            project   = cachmngr.read_project(str(path), strict=strict)
            timestamp = modified_time

        return project, timestamp




#
# Copyright 2018 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Supervisor implementation for controlling lifetime and restart of workers

    This supervisor will not work for distributed workers because it needs to know
    the pid of the worker processes to send abort signal to. 
"""

#
# Client
#

import os
import signal
import asyncio
import zmq.asyncio
import zmq
import traceback
import logging


from .utils import _get_ipc

from collections import namedtuple
from typing import Awaitable, Any

LOGGER=logging.getLogger('SRVLOG')

_Report = namedtuple('_Report',('data'))


class Client:

    def __init__(self) -> None:
        """ Supervised client notifier
        """
        address = _get_ipc('supervisor')

        ctx = zmq.Context.instance()
        self._sock = ctx.socket(zmq.PUSH)
        self._sock.setsockopt(zmq.IMMEDIATE, 1) # Do no queue if no connection
        self._sock.connect(address)
        self._pid = os.getpid()
        self._busy = False

    def _send(self, data: Any ) -> None: 
        try:
            self._sock.send_pyobj((self._pid, data), flags=zmq.DONTWAIT)
        except zmq.ZMQError as err:
            if err.errno != zmq.EAGAIN:
                LOGGER.error("%s (%s)", zmq.strerror(err.errno), err.errno)

    def notify_done(self) -> None:
        """ Send 'ready' notification
        """
        if self._busy:
            self._busy = False
            self._send(b'DONE')

    def notify_busy(self) -> None:
        """ send 'busy' notification
        """
        if not self._busy:
            self._busy = True
            self._send(b'BUSY')

    def close(self) -> None:
        self._sock.close()

    def send_report(self, data: Any) -> None:
        self._send(_Report(data=data))
    


class Supervisor:

    def __init__(self, timeout: int)-> None:
        """ Run supervisor

            :param timeout: timeout delay in seconds
        """
        address = _get_ipc('supervisor')

        ctx = zmq.asyncio.Context.instance()
        self._sock = ctx.socket(zmq.PULL)
        self._sock.setsockopt(zmq.RCVTIMEO, 1000)
        self._sock.bind( address )

        self._timeout = timeout
        self._busy = {}
        self._stopped = True
        self._task = None
        self._reports = {}

    def run(self) -> None:
        self._task = asyncio.ensure_future(self._run_async())

    async def _run_async(self) -> Awaitable[None]:
        """ Run supervisor
        """
        loop = asyncio.get_event_loop()

        def kill(pid:int) -> None:
            del self._busy[pid]
            try:
                os.kill(pid, signal.SIGKILL)
                LOGGER.critical("Killed stalled process %s", pid)
            except ProcessLookupError:
                # Process was already terminated/crashed
                pass

        self._stopped = False

        while not self._stopped:
            try:
                pid, msg = await self._sock.recv_pyobj()
                if msg == b'BUSY':
                    self._busy[pid] = loop.call_later(self._timeout,kill,pid)
                elif msg == b'DONE':
                    try:
                        self._busy.pop(pid).cancel()
                    except KeyError:
                        pass
                elif isinstance(msg, _Report):
                    self._reports[pid] = msg.data
            except zmq.ZMQError as err:
                if err.errno != zmq.EAGAIN:
                    LOGGER.error("%s\n%s", zmq.strerror(err.errno), traceback.format_exc())
            except asyncio.CancelledError:
                raise
            except Exception:
                LOGGER.critical("%s", traceback.format_exc())
                raise

    @property
    def reports(self):
        return list(self._reports.values())

    def num_reports(self) -> int:
        return len(self._reports)

    def clear_reports(self):
        self._reports = {}

    def stop(self) -> None:
        """ Stop the supervisor
        """
        LOGGER.info("Stopping supervisor")
        self._stopped = True
        if self._task and not self._task.cancelled():
            self._task.cancel()
        for th in self._busy.values():
            th.cancel()
        self._busy.clear()





#
# Copyright 2020 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Postgres protocol handler

    See 
      * http://www.bostongis.com/blog/index.php?/archives/271-New-in-QGIS-3.2-Save-Project-to-PostgreSQL.html
      * https://github.com/qgis/QGIS-Enhancement-Proposals/issues/118

    project filename syntax for posgresql

    ```
    postgres://[user[:pass]@]host[:port]/?dbname=X&schema=Y&project=Z
    ```

    How does QGIS store the project?

    QGIS creates a table called qgis_projects in whatever schema you had specified, each project is stored as a separate row.
    
    columns are: 'name', 'metadata', 'content'

    Metadata is a json data that contains the 'last_modified_time' field. 


"""
import logging
import urllib.parse

import psycopg2

from urllib.parse import parse_qs

from typing import Tuple, Optional
from datetime import datetime

from qgis.core import QgsProject
from pyqgisservercontrib.core import componentmanager

LOGGER = logging.getLogger('SRVLOG')

__all__= []

# List of allowed params in secure mode
ALLOWED_SECURE_PARAMS=('service','project','dbname','schema')

#
# Connect to database and check modification time
# Qgis is silent when failing to read a project from database and really 
# makes it hard to debug.
#
# So we check the timestamp ourselve and makes a consistency check to ensure
# That the project has benn read correctly
#

def _check_unsafe_url( insecure: bool, url: urllib.parse.ParseResult ) -> Tuple[str,datetime]:
    """ Check unsafe url
    """
    if insecure:      
        LOGGER.warning("Setting postgres connexion parameters in insecure mode %s", url.geturl())
        params = { k:v[0] for k,v in parse_qs(url.query).items() }
    else:
        # Secure mode: allow only secure parameter
        params = { k:v[0] for k,v in parse_qs(url.query).items() if k in ALLOWED_SECURE_PARAMS }

    # Support "postgres:<project>" short form
    prjname  = params.pop('project', url.path.strip('/'))
    if not prjname:
        LOGGER.error("Postgres handler: Missing project parameter '%s'", url.geturl()) 
        raise FileNotFoundError(url.geturl())

    schema   = params.pop('schema','public')
    database = params.pop('dbname',None)
 
    connexion_params = dict(host=url.hostname,
                            port=url.port,
                            user=url.username,
                            password=url.password,
                            database=database,
                            # Treats remaining params as supported psql client options
                            **params)
    if insecure:
        netloc = url.netloc
    else:
        netloc = f'{url.username}@' if url.username else ''

    # Create secure url
    params = '&'.join('%s=%s' % (k,v) for k,v in params.items())
    params = f"project={prjname}&schema={schema}&{params}"
    if database:
        params = f"{params}&dbname={database}"

    urlstr = f"postgresql://{netloc}/?{params}"

    try:
        LOGGER.debug("**** Postgresql connection params %s", connexion_params)
        conn   = psycopg2.connect(**connexion_params)
        cursor = conn.cursor()
        cursor.execute(f"select metadata from {schema}.qgis_projects where name='{prjname}'")
        if cursor.rowcount <= 0:
            raise FileNotFoundError(url.geturl())
        metadata = cursor.fetchone()[0]
        LOGGER.debug("**** Postgres metadata for '%s': %s", prjname, metadata)
        conn.close()
    except psycopg2.OperationalError as e:
        LOGGER.error("Postgres handler Connection error: %s", str(e))
        raise FileNotFoundError(urlstr)
    except psycopg2.Error as e:
        LOGGER.error("Postgres handler Connection error: %s", str(e))
        raise RuntimeError("Connection failed: %s", urlstr)

    modified_time = datetime.fromisoformat(metadata['last_modified_time'])

    return urlstr, modified_time
    

@componentmanager.register_factory('@3liz.org/cache/protocol-handler;1?scheme=postgres')
class PostgresProtocolHandler:
    """ Handle postgres protocol
    """
    def __init__(self):
        cnf = componentmanager.get_service('@3liz.org/config-service;1')
        self._insecure = cnf.getboolean('projects.cache','insecure', fallback=False)

    def get_project( self, url: urllib.parse.ParseResult, strict: Optional[bool]=None,
                     project: Optional[QgsProject]=None,
                     timestamp: Optional[datetime]=None) -> Tuple[QgsProject, datetime]:
        """ Create or return a project

            .. versionadded:: 1.3.2

            Supports the postgres:///projectname syntax
        """
        urlstr, modified_time = _check_unsafe_url( self._insecure, url )

        if timestamp is None or timestamp < modified_time:
            cachemngr = componentmanager.get_service('@3liz.org/cache-manager;1')
            project   = cachemngr.read_project(urlstr)
            timestamp = modified_time

        return project, timestamp




#
# Copyright 2019 3liz
# Author: David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


""" Define middleware filters

    Declare filters as

    @blockingfilter(pri=100, uri="/ows/")
    def myfilter1( handller: tornado.web.RequestHandler, *args ) -> None:
        ...

    @asyncfilter(pri=100, uri="/ows/")
    async def myfilter2( handller: tornado.web.RequestHandler, *args) -> None:
        ...


    Filters are applied before sending the request, a filter can return response or 
    raise exception based on authorization rules.

    Registering filters:

    Modules implementing filters must register their filters using setup.py entry_points with the 
    key 'pyqgisserver_filters'

    def register_filters():
        # Do initialization stuff
        return [ myfilter1, myfilter2 ]
"""

import logging
import tornado.web

from typing import Coroutine, Callable, Awaitable, Optional

LOGGER = logging.getLogger('SRVLOG')


class ServerFilter:
    pass


class asyncfilter(ServerFilter):
    """ Decorator for asynchronous request filter

        Assume that the function return a  coroutine
    """

    def __init__(self, pri=0, uri=""):
        self.pri = pri
        self.uri = uri

    def __call__(self, fn: Callable[[tornado.web.RequestHandler], Awaitable[None]]) -> 'asyncfilter':
        self.fn = fn
        return self

    def apply( self, handler: tornado.web.RequestHandler ) -> Coroutine:
        return self.fn(handler)


class blockingfilter(ServerFilter):
    """ Decorator for synchronous request filter
    """

    def __init__(self, pri=0, uri=""):
        self.pri = pri
        self.uri = uri

    def __call__(self, fn):
        self.fn = fn
        return self

    async def apply( self, handler: tornado.web.RequestHandler ) -> Optional[str]:
        return self.fn(handler)




#
# Copyright 2018-2019 3liz
# Author: David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Qgis server handler
"""
import logging
from time import time

from ..logger import log_rrequest
from ..zeromq.client import RequestTimeoutError, RequestGatewayError, AsyncClient
from ..monitor import Monitor

from .basehandler import BaseHandler

from typing import Optional, Awaitable, List

LOGGER = logging.getLogger('SRVLOG')


class OwsHandler(BaseHandler):

    """ Proxy to Qgis 0MQ worker
    """
    def initialize(self, client: AsyncClient, timeout: int, 
                   monitor: Optional[Monitor]=None, 
                   http_proxy: bool=False,
                   allowed_hdrs: List[str]=[] ) -> None:

        super().initialize()

        self._client       = client
        self._timeout      = timeout
        self._monitor      = monitor
        self._proxy        = http_proxy
        self._stats        = self.application.stats
        self._allowed_hdrs = allowed_hdrs

        self.ogc_scheme   = None
        
    async def handle_request(self, method: str, endpoint: Optional[str]=None) -> Awaitable[None]:
        reqtime = time()
        try:
            proxy_url = self.proxy_url(self._proxy, endpoint)

            delta = None
            project_path = self.get_argument('MAP',default=None)
            query        = self.encode_arguments()

            headers = {}

            data = self.request.body

            if project_path:
                headers['X-Map-Location']=project_path 
            if proxy_url: 
                headers['X-Forwarded-Url']=proxy_url

            if self.ogc_scheme:
                headers['X-Ogc-Scheme'] = self.ogc_scheme

            def copy_headers(pats):
                headers.update((k,v) for k,v in self.request.headers.items() if \
                               any(map(k.upper().startswith,pats)))

            # Copy custom Qgis/Forwarded headers
            # see https://github.com/qgis/QGIS/pull/41333
            copy_headers(self._allowed_hdrs)

            if self.get_argument('SERVICE', default=None) and  self.has_body_arguments:
                # Do not let qgis server handle url encoded parameters
                data = None
                if method == 'POST':
                    method = 'GET'

            self._stats.num_requests +=1

            response = await self._client.fetch(query=query, method=method, 
                                                headers=headers, data=data,
                                                timeout=self._timeout)
            status = response.status
            hdrs   = response.headers
            delta  = time() - reqtime

            log_rrequest(proxy_url, status, method, query, delta, hdrs)
           
            # Send response
            for k,v in hdrs.items():
                self.set_header(k,v)

            # Send CORS Header
            self.set_access_control_headers()

            if status == 206:
                # Partial response
                self.set_status(200)
                self.write(response.data)
                await self.flush()
                async for chunk in self._client.fetch_more(response, timeout=self._timeout):
                    self.write(chunk)
                    await self.flush()
                delta = time() - reqtime
            elif status == 509:
                self.send_error(status, reason="Server busy, please retry later")
            else:
                self.set_status(status)
                self.write(response.data)

        except RequestTimeoutError:
            status = 504
            delta = time() - reqtime
            # Log the request with status code 499 indicating
            # that the request has not returned
            log_rrequest(proxy_url, 499, method, query, delta, {})
            self.send_error(status, reason="Request timeout error")
        except RequestGatewayError:
            status = 502
            delta = time() - reqtime
            # Log the request 
            log_rrequest(proxy_url, 499, method, query, delta, {})
            self.send_error(status, reason="Backend request error")

        if status >= 500:
            self._stats.num_errors +=1

        if self._monitor:
            self._monitor.emit( status, self.request.arguments,  delta, 
                                meta=self.request.headers)

    async def get(self, endpoint: Optional[str]=None) -> Awaitable[None]:
        """ Handle Get method
        """
        await self.handle_request('GET', endpoint)
          
    async def post(self, endpoint: Optional[str]=None) -> Awaitable[None]:
        """ Handle Post method
        """
        await self.handle_request('POST', endpoint)
        
    def options(self, endpoint: Optional[str]=None) -> None:
        """ Implement OPTION for validating CORS
        """
        self.set_option_headers('GET, POST, OPTIONS')


class _FilterHandlerMixIn:
    """ Handle filter handlers
    """
    def initialize(self, filters: Optional[List]=None, **kwargs) -> None:
        super().initialize(**kwargs)
        
        self._filters = filters or []

    async def prepare(self) -> Awaitable[None]:
        # Handle filters
        super().prepare()
        for filt in self._filters:
            await filt.apply( self )


class OwsFilterHandler(_FilterHandlerMixIn,OwsHandler):
    pass


class OwsApiHandler(OwsHandler):
    """ Handle Qgis api
    """
    def initialize( self, *args, **kwargs ) -> None:
        super().initialize(*args, **kwargs )
        self.ogc_scheme = 'OAF'

    async def get(self, endpoint: Optional[str]=None) -> Awaitable[None]:
        """ Fix issue with the landing page api when not 
            specifying index.html.
        """
        await super().get(endpoint)
 
    async def delete(self, endpoint: Optional[str]=None) -> Awaitable[None]:
        await self.handle_request('DELETE', endpoint)

    async def head(self, endpoint: Optional[str]=None) -> Awaitable[None]:
        await self.handle_request('HEAD', endpoint)

    async def put(self, endpoint: Optional[str]=None) -> Awaitable[None]:
        await self.handle_request('PUT', endpoint)

    async def patch(self, endpoint: Optional[str]=None) -> Awaitable[None]:
        await self.handle_request('PATCH', endpoint)

    async def options(self, endpoint: Optional[str]=None) -> Awaitable[None]:
        await self.handle_request('OPTIONS', endpoint)


class OwsApiFilterHandler(_FilterHandlerMixIn, OwsApiHandler):
    pass




#
# Copyright 2021 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from pyqgisserver.management.apis import plugins, cache


def register_management_apis( serverIface ):
    """ Return management Qgis server apis
    """
    plugins.register( serverIface )
    cache.register( serverIface )    




#
# Copyright 2020 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

""" Cache manager for Qgis Projects
"""

import logging
import urllib.parse

from urllib.parse import urlparse, urljoin, parse_qs
from typing import Tuple, Optional, Sequence
from collections import namedtuple
from pathlib import Path
from datetime import datetime

from ..utils.lru import lrucache
from ..config import confservice

from qgis.PyQt.QtCore import Qt 
from qgis.core import QgsProjectBadLayerHandler, QgsProject, QgsMapLayer
from qgis.server import QgsServerProjectUtils

from pyqgisservercontrib.core import componentmanager

# Import default handlers for auto-registration
from .handlers import * # noqa: F403,F401

LOGGER = logging.getLogger('SRVLOG')


class StrictCheckingError(Exception):
    pass


class PathNotAllowedError(Exception):
    pass


class UnreadableResourceError(Exception):
    """ Indicates that the  ressource exists but is not readable
    """
    pass


CacheDetails = namedtuple("CacheDetails",('project','timestamp'))


CACHE_MANAGER_CONTRACTID = '@3liz.org/cache-manager;1'


def _merge_qs( query1: str, query2: str ) -> str:
    """ Merge query1 with query2 but coerce values
        from query1 
    """
    params_1 = parse_qs(query1)
    params_2 = parse_qs(query2)
    params_2.update(params_1)
    return '&'.join('%s=%s' % (k,v[0]) for k,v in params_2.items())


@componentmanager.register_factory(CACHE_MANAGER_CONTRACTID)
class QgsCacheManager:
    """ Handle Qgis project cache 
    """

    StrictCheckingError=StrictCheckingError
    PathNotAllowedError=PathNotAllowedError
    UnreadableResourceError=UnreadableResourceError

    def __init__(self) -> None:
        """ Initialize cache

            :param size: size of the lru cache
        """
        cnf = confservice['projects.cache']

        size = cnf.getint('size')

        self._create_project = QgsProject
        self._cache = lrucache(size)
        self._strict_check         = cnf.getboolean('strict_check')
        self._trust_layer_metadata = cnf.getboolean('trust_layer_metadata')
        self._disable_getprint     = cnf.getboolean('disable_getprint')
        self._disable_owsurls      = cnf.getboolean('disable_owsurls', fallback=False)
        self._aliases = {}
        self._default_scheme = cnf.get('default_handler',fallback='file')

        # Set the base url for file protocol
        self._aliases['file'] = 'file:///%s/' % cnf.get('rootdir').strip('/')

        # Load protocol handlers
        componentmanager.register_entrypoints('qgssrv_contrib_protocol_handler')

    def clear(self) -> None:
        """ Clear the whole cache
        """
        self._cache.clear()

    def items(self) -> Sequence[Tuple[str,CacheDetails]]:
        return self._cache.items()

    def remove_entry(self, key: str) -> None:
        """ Remove cache entry
        """
        del self._cache[key]

    def resolve_alias(self, key: str ) -> urllib.parse.ParseResult:
        """ Resolve scheme from configuration variables
        """
        url    = urlparse(key)
        scheme = url.scheme or self._default_scheme
        LOGGER.debug("Resolving '%s' protocol", scheme)
        baseurl = self._aliases.get(scheme)
        if not baseurl:
            try:
                # Check for user-defined scheme
                baseurl = confservice.get('projects.schemes',scheme.replace('-','_').lower())
            except KeyError:
                pass
            else:
                LOGGER.info("Scheme '%s' aliased to %s", scheme, baseurl)
                self._aliases[scheme] = baseurl
        if baseurl:
            if '{path}' in baseurl:
                url = urlparse(baseurl.format(path=url.path))
            else:
                baseurl = urlparse(baseurl)
                # Build a new query from coercing with base url params 
                query = _merge_qs(baseurl.query, url.query)
                # XXX Note that the path of the base url must be terminated by '/'
                # otherwise urljoin() will replace the base name - may be not what we want
                url = urlparse(urljoin(baseurl.geturl(),url.path+'?'+query))
                # Make sure that the result url path is relative to base url
                try:
                    # Ensure that if an absolute path is given, we may extract
                    # a relative path to the base url - note that the base url may
                    # not have a leading '/'
                    Path(url.path).relative_to(Path('/') / baseurl.path)
                except ValueError:
                    LOGGER.error("The path '%s' is outside base path '%s'", url.path, baseurl.path)
                    raise PathNotAllowedError()                    

        return url

    def get_project(self, key: str, strict: Optional[bool]=None) -> Tuple[QgsProject,datetime,bool]:
        """ Load project 
        """
        url = self.resolve_alias(key)
    
        scheme = url.scheme or self._default_scheme
        # Retrieve the protocol-handler
        try:
            store = componentmanager.get_service('@3liz.org/cache/protocol-handler;1?scheme=%s' % scheme)
        except componentmanager.FactoryNotFoundError:
            LOGGER.error("No protocol handler found for %s", scheme) 
            raise FileNotFoundError(key)

        # Get details for the project
        details = self._cache.peek(key)
        if details is not None:
            project, timestamp  = store.get_project( url, strict=strict, **details._asdict())
            needupdate = timestamp != details.timestamp
        else:
            project, timestamp = store.get_project(url, strict=strict)
            needupdate = True

        return project, timestamp, needupdate
        
    def update_entry(self, key: str) -> bool:
        """ Update the cache

            :param key: The key of the entry to update
            :return: true if the entry has been updated
        """
        project, timestamp, updated = self.get_project(key)
        self._cache[key] = CacheDetails(project, timestamp)
        return updated

    def peek(self, key: str) -> Optional[CacheDetails]:
        """ Return entry if it exists
        """
        return self._cache.peek(key)

    def lookup(self, key: str) -> Tuple[QgsProject, bool]:
        """ Lookup entry from key
        """
        updated = self.update_entry(key)
        return self._cache[key].project, updated

    def prepare_project(self, project: QgsProject ) -> None:
        """ Set project configuration
        """
        if self._disable_owsurls:
            # Disable ows urls defined in project
            # May be needed because it overrides 
            # any proxy settings
            project.writeEntry("WMSUrl","/", "")
            project.writeEntry("WFSUrl","/", "")
            project.writeEntry("WCSUrl","/", "")

    def read_project(self, path: str, strict: Optional[bool]=None) -> QgsProject:
        """ Read project from path

            May be used by protocol-handlers to instanciate project
            from path.
        """
        LOGGER.debug("Reading Qgis project %s", path)
        project = self._create_project()

        readflags = QgsProject.ReadFlags()
        if self._trust_layer_metadata:
            readflags |= QgsProject.FlagTrustLayerMetadata
        if self._disable_getprint:
            readflags |= QgsProject.FlagDontLoadLayouts 
        badlayerh = BadLayerHandler()
        project.setBadLayerHandler(badlayerh)
        project.read(path,  readflags)

        strict = self._strict_check if strict is None else strict
        if strict and not badlayerh.validateLayers(project):
            raise StrictCheckingError

        self.prepare_project(project)
        return project


class BadLayerHandler(QgsProjectBadLayerHandler):

    def __init__(self):
        super().__init__()
        self.badLayerNames = set()

    def handleBadLayers( self, layers ) -> None:
        """ See https://qgis.org/pyqgis/3.0/core/Project/QgsProjectBadLayerHandler.html
        """
        super().handleBadLayers( layers )

        nameElements = (lyr.firstChildElement("layername") for lyr in layers if lyr)
        self.badLayerNames = set(elem.text() for elem in nameElements if elem)

    def validateLayers( self, project: QgsProject ) -> bool:
        """ Check layers
            
            If layers are excluded do not count them as bad layers
            see https://github.com/qgis/QGIS/pull/33668
        """
        if self.badLayerNames:
            LOGGER.error("Found bad layers: %s", self.badLayerNames)
            restricteds = set(QgsServerProjectUtils.wmsRestrictedLayers(project))
            return self.badLayerNames.issubset(restricteds)
        return True


def get_cacheservice() -> QgsCacheManager:
    return componentmanager.get_service(CACHE_MANAGER_CONTRACTID)


def preload_projects_file( path: Path, cacheservice: QgsCacheManager ) ->  int:
    """ Preload projects from configuration file
    """
    conf_file = Path(path)
    if not conf_file.exists():
        LOGGER.error("%s file do not exists, ignoring preload config", path)
        return 0

    # No point to preload more files than the cache size
    maxfiles = confservice['projects.cache'].getint('size')
    loaded_so_far = 0
    
    # Read the projects, strip comments 
    with conf_file.open() as fp:
        for p in filter(None,(line.strip('\n ').partition('#')[0] for line in fp.readlines())):
            p = p.strip(' ')
            try:
                project, updated = cacheservice.lookup(p)
            except StrictCheckingError:
                LOGGER.error("Preload: '%s' as invalid layers - strict mode on" , p)
            except PathNotAllowedError:
                LOGGER.error("Preload: '%s' path not allowed", p)
            except FileNotFoundError:
                LOGGER.error("Preload: '%s' not found", p)
            else:
                LOGGER.info("Preload: '%s' loaded", p)
                loaded_so_far += 1
            if loaded_so_far >= maxfiles:
                LOGGER.warning("Preload: cache size reached")
                break
    return loaded_so_far


def preload_projects() -> None:
    """ Preload projects in cache
    """
    confpath = confservice['projects.cache'].get('preload_config', fallback=None)
    if not confpath:
        return

    preload_projects_file( confpath, get_cacheservice() )


def get_project_summary( key: str, project: QgsProject ):
    """ Return json summary for cached project
    """
    def layer_summary( layer_id: str, layer: QgsMapLayer ):
        return dict(
            id=layer_id,
            name=layer.name(),
            source=layer.publicSource(),
            crs=layer.crs().userFriendlyIdentifier(),
            valid=layer.isValid(),
            spatial=layer.isSpatial(),
        )

    layers = [layer_summary(idstr,l) for (idstr,l) in project.mapLayers().items()]

    return dict(
        cache_key=key,
        filename=project.fileName(),
        bad_layers_count=sum(1 for ls in layers if not ls['valid']),
        layers=layers,
        crs=project.crs().userFriendlyIdentifier(),
        last_modified=project.lastModified().toString(Qt.ISODate)
    )



#
# Copyright 2021 3liz
# Author David Marteau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
    Wrapper around more tornado-like request handler for
    implementing Qgis server API
"""
import logging
import sys
import json
import traceback

from qgis.PyQt.QtCore import QRegularExpression, QUrl

from qgis.server import (
    QgsServerOgcApi,
    QgsServerOgcApiHandler,
    QgsServerRequest,
)

from tornado.web import HTTPError # noqa F401
from tornado import httputil

from typing import Any, Optional, Union, Tuple, Type, List

LOGGER = logging.getLogger('SRVLOG')


class RequestHandler:
  
    def __init__(self, context):
        self._context  = context
        self._response = context.response()
        self._request  = context.request()
        self._finished = False

    def prepare(self):
        """
        """

    def public_url(self, path="", rootpath: Optional[str]=None) -> str:
        """ Return the public base url
        """
        url = self._request.originalUrl()
        rootpath = rootpath or self._context.apiRootPath() 
        public_url = f"{url.scheme()}://{url.authority()}{rootpath}{path}"
        return public_url

    def finish(self, chunk: Optional[Union[str, bytes, dict]] = None) -> None:
        """
        """
        if self._finished:
            raise RuntimeError("finish() called twice")

        if chunk is not None:
            self.write(chunk)

        self._finished = True
        self._response.finish()
        
    def write(self, chunk: Union[str, bytes, dict]) -> None: 
        """
        """
        if not isinstance(chunk, (bytes, str, dict)):
            raise TypeError("write() only accepts bytes, unicode, and dict objects")
        if isinstance(chunk, dict):
            chunk = json.dumps(chunk, sort_keys=True)
        self.set_header('Content-Type', 'application/json;charset=utf-8')
        self._response.write(chunk) 

    def set_status(self, status_code: int, reason: Optional[str]=None) -> None:
        """
        """
        self._response.setStatusCode(status_code)
        if reason is not None:
            self._reason = reason
        else:
            self._reason = httputil.responses.get(status_code, "Unknown")

    def send_error(self, status_code: int = 500, **kwargs: Any) -> None:
        """
        """
        self._response.clear()
        reason = kwargs.get("reason")
        if "exc_info" in kwargs:
            exception = kwargs["exc_info"][1]
            if isinstance(exception, HTTPError) and exception.reason:
                reason = exception.reason
        self.set_status(status_code, reason=reason)
        self.write(dict(status="error" if status_code != 200 else "ok",
                        httpcode = status_code,
                        error    = { "message": self._reason }))
        if not self._finished:
            self.finish()
        
    def set_header(self, name: str, value: str) -> None:
        """
        """
        self._response.setHeader(name,value)

    def _unimplemented_method(self, *args: str, **kwargs: str) -> None:
        raise HTTPError(405)

    head   = _unimplemented_method
    get    = _unimplemented_method  
    post   = _unimplemented_method  
    delete = _unimplemented_method  
    patch  = _unimplemented_method 
    put    = _unimplemented_method  

    METHODS = {
        QgsServerRequest.HeadMethod  : 'head',
        QgsServerRequest.PutMethod   : 'put',
        QgsServerRequest.GetMethod   : 'get',
        QgsServerRequest.PostMethod  : 'post',
        QgsServerRequest.PatchMethod : 'patch',
        QgsServerRequest.DeleteMethod: 'delete',
    }

    def _execute(self, values):
        """ Execute the request
        """
        try:
            method = self._request.method()
            if method not in self.METHODS:
                raise HTTPError(405)

            self.prepare()
            if self._finished:
                return

            self.set_status(200)

            method = getattr(self, self.METHODS[method])            
            method( **values )

            if not self._finished:
                self.finish()
        except Exception as e:
            if self._finished:
                # Nothing to send, but log for debugging purpose
                LOGGER.error(traceback.format_exc())
                return
            if isinstance(e, HTTPError):
                self.send_error(e.status_code, exc_info=sys.exc_info())
            else:
                LOGGER.error(traceback.format_exc())
                self.send_error(500, exc_info=sys.exc_info())


class RequestHandlerDelegate(QgsServerOgcApiHandler):
    """ Delegate request to handler
    """

    # XXX We need to preserve instances from garbage
    # collection 
    __instances = []

    def __init__(self, path: str, handler: Type[RequestHandler], 
                 content_types=[QgsServerOgcApi.JSON,]):

        super().__init__()
        if content_types:
            self.setContentTypes(content_types)
        self._path = QRegularExpression(path)
        self._name = handler.__name__
        self._handler = handler

        self.__instances.append(self)

    def path(self):
        return self._path

    def linkType(self):
        return QgsServerOgcApi.data

    def operationId(self):
        return f"PyQgisServer{self._name}Management"

    def summary(self):
        return f"PyQgisServerManagement {self._name}  Service"

    def description(self):
        return f"PyQgisServerManagement {self._name } Service "

    def linkTitle(self):
        return f"PyQgisServerManagement {self._name } Service "

    def templatePath(self, context):
        # No templates!
        return ''

    def parameters(self, context):
        return []

    def handleRequest(self, context):
        """ 
        """
        handler = self._handler(context)
        handler._execute(self.values(context))


class _ServerApi(QgsServerOgcApi):

    __instances = []

    # See above
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)

        self.__instances.append(self)

    def accept(self, url: QUrl) -> bool:
        """ Override the api to actually match the rootpath
        """
        return url.path().startswith( self.rootPath() )



def register_handlers(serverIface, rootpath: str, name: str, handlers: List[Tuple[str,Type[RequestHandler]]],
                      descripton: Optional[str]=None,
                      version: Optional[str]=None) -> None:

    api = _ServerApi(serverIface,rootpath, name, descripton, version)
    for (path,handler) in handlers:
        api.registerHandler(RequestHandlerDelegate(path,handler))

    serverIface.serviceRegistry().registerApi(api)






from setuptools import setup, find_namespace_packages
import os

def parse_requirements( filename ):
    with open( filename ) as fp:
        return list(filter(None, (r.strip('\n ').partition('#')[0] for r in fp.readlines())))

def get_version():
    with open('VERSION') as fp:
        return fp.read().strip()

VERSION = get_version()
DESCRIPTION = ('Py-Qgis-Server is an implementation of the OWS standards '
               'standard from the Open Geospatial Consortium.'
               'Py-Qgis-Server written in Python and built on top of Qgis server'
               'implementation')
KEYWORDS = 'QGIS OWS OGC'
kwargs = {}

with open('README.md') as f:
    kwargs['long_description'] = f.read()

# Parse requirement file and transform it to setuptools requirements'''
requirements = 'requirements.txt'
if os.path.exists(requirements):
    kwargs['install_requires']=parse_requirements(requirements)

setup(
    name='py-qgis-server',
    version=VERSION,
    author='3Liz',
    author_email='david.marteau@3liz.com',
    maintainer='David Marteau',
    maintainer_email='dmarteau@3liz.com',
    long_description_content_type="text/markdown",
    description=DESCRIPTION,
    keywords=KEYWORDS,
    url='https://github.com/3liz/py-qgis-server',
    packages=find_namespace_packages(include=['pyqgisserver','pyqgisserver.*',
                                              'pyqgisservercontrib.*']),
    entry_points={
        'console_scripts': [
            'qgisserver = pyqgisserver.server:main',
            'qgisserver-worker = pyqgisserver.qgsworker:main',
        ],
    },
    # Add manifest to main package
    include_package_data=True,
    package_data={"pyqgisserver": ['build.manifest'] },
    python_requires='>=3.6',
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    **kwargs
)
